"""Compose nodes with specific agent and context.

This module provides the `compose` function for executing nodes with dependency injection.

Key features:
- Execute functions as nodes with automatic dependency resolution
- Work with Node classes and ready-made Composable objects
- Integration with `Context` for access to API, events, data, and other nodes

```python
# Executing a function node
async def process_data(user_id: int, api: API) -> str:
    user = await api.get_chat(user_id)
    return f"Hello, {user.first_name}!"

async def handler(context: Context):
    async with compose(process_data, context) as result:
        match result:
            case Ok(value):
                print(value)  # "Hello, John!"
            case Error(error):
                print(f"Error: {error}")

# Executing a node with additional dependencies
from telegrinder.node import per_call, scalar_node

@per_call
@scalar_node
class Database:
    @classmethod
    def __compose__(cls) -> "Database":
        return cls()

async def get_user_data(db: Database, user_id: int) -> dict[str, Any]:
    # Use db to fetch data
    return {"id": user_id, "name": "User"}

async def handler(context: Context):
    async with compose(get_user_data, context) as result:
        user_data = result.unwrap()

# Passing additional root dependencies
async def handler(context: Context):
    custom_data = {"custom_key": "custom_value"}

    async with compose(
        my_node,
        context,
        roots={CustomType: custom_data}
    ) as result:
        # roots allows injecting additional dependencies
        pass
```

The `compose` function returns an async context manager that:

- Creates a local scope for node composing
- Automatically resolves all node dependencies
- Returns `Result[T, NodeError]` with the composed result
"""

import dataclasses
import typing
from contextlib import asynccontextmanager
from functools import lru_cache

from kungfu.library.monad.result import Error, Ok, Result
from nodnod.agent.base import Agent
from nodnod.agent.event_loop.agent import EventLoopAgent
from nodnod.error import NodeError
from nodnod.interface import create_agent_from_node, create_node_from_function, inject_internals
from nodnod.interface.is_node import is_node
from nodnod.interface.node_from_function import Externals
from nodnod.node import Node
from nodnod.scope import Scope
from nodnod.utils.is_type import is_type

from telegrinder.node.scope import TELEGRINDER_CONTEXT, MappedScopes, NodeScope
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.tools.magic.function import Function

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.context import Context

type _Composable[T: Agent] = Function[..., typing.Any] | type[Node[typing.Any, typing.Any]] | Composable[T]

NODE_GLOBAL_SCOPE: typing.Final = TELEGRINDER_CONTEXT.node_global_scope


@lru_cache(maxsize=1024 * 4)
def create_composable[T: Agent](
    node_or_function: type[Node[typing.Any, typing.Any]] | Function[..., typing.Any],
    /,
    *,
    agent_cls: type[T] = EventLoopAgent,
) -> Composable[T]:
    if not isinstance(node_or_function, type):
        node_or_function = create_node_from_function(node_or_function)
    return Composable(node_or_function, create_agent_from_node(node_or_function, agent_cls=agent_cls))


@asynccontextmanager
async def run_agent(
    agent: Agent,
    context: Context,
    *,
    roots: dict[type[typing.Any], typing.Any] | None = None,
    per_event_scope: Scope | None = None,
) -> typing.AsyncGenerator[Result[Scope, NodeError]]:
    event_scope = per_event_scope if per_event_scope is not None else context.per_event_scope
    mapped_scopes = MappedScopes(global_scope=NODE_GLOBAL_SCOPE, per_event_scope=event_scope)
    internals = {type(context): context, Externals: context}

    async with event_scope.create_child(detail=NodeScope.PER_CALL) as local_scope:
        try:
            inject_internals(local_scope, internals | (roots or {}))
            await maybe_awaitable(agent.run(local_scope, mapped_scopes))
            yield Ok(local_scope.merge())
        except NodeError as error:
            yield Error(error)


@typing.overload
def compose[R](
    function: Function[..., R],
    /,
    context: Context,
    *,
    per_event_scope: Scope | None = None,
    agent_cls: type[Agent] | None = None,
    roots: dict[type[typing.Any], typing.Any] | None = None,
) -> typing.AsyncContextManager[Result[R, NodeError], None]: ...


@typing.overload
def compose[T: Agent](
    composable: Composable[T],
    /,
    context: Context,
    *,
    per_event_scope: Scope | None = None,
    roots: dict[type[typing.Any], typing.Any] | None = None,
) -> typing.AsyncContextManager[Result[typing.Any, NodeError], None]: ...


@typing.overload
def compose[T](
    node: type[Node[T, typing.Any]],
    /,
    context: Context,
    *,
    agent: Agent,
    per_event_scope: Scope | None = None,
    roots: dict[type[typing.Any], typing.Any] | None = None,
) -> typing.AsyncContextManager[Result[T, NodeError], None]: ...


@asynccontextmanager
async def compose[T = typing.Any](
    node: _Composable[Agent],
    context: Context,
    *,
    per_event_scope: Scope | None = None,
    agent: Agent | None = None,
    agent_cls: type[Agent] | None = None,
    roots: dict[type[typing.Any], typing.Any] | None = None,
) -> typing.AsyncGenerator[Result[T, NodeError], None]:
    composable = None

    if isinstance(node, Composable):
        composable = node
    elif not is_node(node):
        composable = create_composable(node, agent_cls=agent_cls or EventLoopAgent)

    if composable is not None:
        node, agent = composable.node, typing.cast("Agent", composable.agent)
    elif not is_type(node, Node):
        raise TypeError("Compose requires function, node, or composable.")

    if agent is None:
        raise ValueError("Agent is required.")

    async with run_agent(agent, context, roots=roots, per_event_scope=per_event_scope) as result:
        yield result.map(lambda scope: scope[node].value)


@dataclasses.dataclass(frozen=True, slots=True)
class Composable[T: Agent = Agent]:
    node: type[Node[typing.Any, typing.Any]]
    agent: T


__all__ = ("Composable", "compose", "create_composable", "run_agent")



from telegrinder.client.abc import ABCClient, Response
from telegrinder.client.form_data import MultipartBuilderProto, encode_form_data
from telegrinder.client.rnet import RnetClient

__all__ = (
    "ABCClient",
    "MultipartBuilderProto",
    "Response",
    "RnetClient",
    "encode_form_data",
)



import dataclasses
import typing

from kungfu.library.monad.option import NOTHING, Nothing, Option
from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node
from nodnod.node import Node

import telegrinder.types
from telegrinder.bot.cute_types.message import MessageCute

type AttachmentType = typing.Literal[
    "animation",
    "audio",
    "document",
    "photo",
    "poll",
    "video",
    "video_note",
    "voice",
    "successful_payment",
]

ATTACHMENT_TYPES: typing.Final[tuple[AttachmentType, ...]] = typing.get_args(AttachmentType.__value__)


@dataclasses.dataclass
class Attachment(Node):
    attachment_type: AttachmentType

    animation: Option[telegrinder.types.Animation] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    audio: Option[telegrinder.types.Audio] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    document: Option[telegrinder.types.Document] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    photo: Option[list[telegrinder.types.PhotoSize]] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    poll: Option[telegrinder.types.Poll] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    voice: Option[telegrinder.types.Voice] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    video: Option[telegrinder.types.Video] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    video_note: Option[telegrinder.types.VideoNote] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    successful_payment: Option[telegrinder.types.SuccessfulPayment] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )

    @classmethod
    def __compose__(cls, message: MessageCute) -> typing.Self:
        for attachment_type in ATTACHMENT_TYPES:
            attachment = getattr(message, attachment_type, NOTHING)

            if attachment:
                return cls(attachment_type, **{attachment_type: attachment})

        raise NodeError("No attachment found in message.")


@dataclasses.dataclass
class Photo(Node):
    sizes: list[telegrinder.types.PhotoSize]

    @classmethod
    def __compose__(cls, attachment: Attachment) -> typing.Self:
        return cls(attachment.photo.expect(NodeError("Attachment is not a photo.")))


@scalar_node
class Video:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Video:
        return attachment.video.expect(NodeError("Attachment is not a video."))


@scalar_node
class VideoNote:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.VideoNote:
        return attachment.video_note.expect(NodeError("Attachment is not a video note."))


@scalar_node
class Audio:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Audio:
        return attachment.audio.expect(NodeError("Attachment is not an audio."))


@scalar_node
class Animation:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Animation:
        return attachment.animation.expect(NodeError("Attachment is not an animation."))


@scalar_node
class Voice:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Voice:
        return attachment.voice.expect(NodeError("Attachment is not a voice."))


@scalar_node
class Document:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Document:
        return attachment.document.expect(NodeError("Attachment is not a document."))


@scalar_node
class Poll:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.Poll:
        return attachment.poll.expect(NodeError("Attachment is not a poll."))


@scalar_node
class SuccessfulPayment:
    @classmethod
    def __compose__(cls, attachment: Attachment) -> telegrinder.types.SuccessfulPayment:
        return attachment.successful_payment.expect(NodeError("Attachment is not a successful payment."))


@dataclasses.dataclass
class MediaGroup(Node):
    id: str
    items: list[MessageCute]

    @classmethod
    def __compose__(cls, message: MessageCute) -> typing.Self:
        return cls(
            id=message.media_group_id.expect(NodeError("No media group id.")),
            items=message.media_group_messages.expect(NodeError("No messages collected for media group.")),
        )


__all__ = (
    "Animation",
    "Attachment",
    "Audio",
    "Document",
    "MediaGroup",
    "Photo",
    "Poll",
    "SuccessfulPayment",
    "Video",
    "VideoNote",
    "Voice",
)



import asyncio
import dataclasses
import datetime
import typing
from contextlib import asynccontextmanager

from telegrinder.modules import logger
from telegrinder.tools.aio import run_task
from telegrinder.tools.fullname import fullname

type CoroutineTask[T] = typing.Coroutine[typing.Any, typing.Any, T]
type CoroutineFunc[**P, T] = typing.Callable[P, CoroutineTask[T]]
type Task[**P, T] = CoroutineFunc[P, T] | CoroutineTask[T] | DelayedTask[P]


def to_coroutine_task[T](task: Task[..., T], /) -> CoroutineTask[T]:
    if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):
        task = task()
    elif not asyncio.iscoroutine(task):
        raise TypeError("Task should be coroutine or coroutine function.")
    return task


@dataclasses.dataclass
class DelayedTask[**P]:
    _cancelled: bool = dataclasses.field(default=False, init=False, repr=False)
    _event: asyncio.Event = dataclasses.field(default_factory=asyncio.Event, init=False, repr=False)
    _timer: asyncio.TimerHandle | None = dataclasses.field(default=None, init=False, repr=False)

    function: typing.Callable[P, CoroutineTask[typing.Any]]
    seconds: float | datetime.timedelta
    repeat: bool = dataclasses.field(default=False, kw_only=True)

    def __post_init__(self) -> None:
        self.function.cancel = self.cancel

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        stopped = False

        while not stopped and not self.is_cancelled:
            self.start_timer()
            await self._event.wait()

            try:
                await self.function(*args, **kwargs)
            except Exception:
                await logger.aexception(
                    "Delayed task `{}` failed with exception, traceback message below:",
                    fullname(self.function),
                )
            finally:
                self._event.clear()
                self._timer = None

                if not self.repeat:
                    stopped = True

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled

    @property
    def delay(self) -> float:
        return float(self.seconds) if isinstance(self.seconds, int | float) else self.seconds.total_seconds()

    def start_timer(self) -> None:
        if self._timer is None:
            self._timer = asyncio.get_running_loop().call_later(
                self.delay,
                callback=lambda: (self._event.set() if not self._event.is_set() else None),
            )

    def cancel(self) -> bool:
        if self._cancelled:
            return True

        self._cancelled = True

        if self._timer is None or self._timer.cancelled():
            self._timer = None
            return False

        self._timer.cancel()
        self._timer = None

        for future in self._event._waiters:
            if not future.cancelled():
                future.cancel()

        return True


@dataclasses.dataclass(kw_only=True, slots=True, repr=False)
class Lifespan:
    _started: bool = dataclasses.field(default=False, init=False)
    _lifespan_context: typing.AsyncContextManager[typing.Any] | None = dataclasses.field(default=None, init=False)
    lifespan_function: typing.Callable[[], typing.AsyncContextManager[typing.Any]] | None = dataclasses.field(
        default=None
    )
    startup_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])
    shutdown_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])

    def __repr__(self) -> str:
        return "<{}: started={}>".format(fullname(self), self._started)

    def __add__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.__class__(
            startup_tasks=self.startup_tasks + other.startup_tasks,
            shutdown_tasks=self.shutdown_tasks + other.shutdown_tasks,
        )

    def __iadd__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__class__):
            return NotImplemented

        self.startup_tasks.extend(other.startup_tasks)
        self.shutdown_tasks.extend(other.shutdown_tasks)
        return self

    async def __aenter__(self) -> None:
        await self._start()

    async def __aexit__(
        self,
        exc_type: typing.Any | None,
        exc_value: typing.Any | None,
        exc_tb: typing.Any | None,
    ) -> None:
        await self._shutdown(exc_type, exc_value, exc_tb)

    def __call__[Function: typing.Callable[[], typing.AsyncGenerator[typing.Any, None]]](
        self,
        func: Function,
        /,
    ) -> Function:
        self.lifespan_function = asynccontextmanager(func)
        return func

    @property
    def started(self) -> bool:
        return self._started

    @staticmethod
    async def _run_tasks(tasks: list[CoroutineTask[typing.Any]], /) -> None:
        while tasks:
            await tasks.pop(0)

    async def _start(self) -> None:
        if not self._started:
            await logger.adebug("Running lifespan startup tasks")
            self._started = True

            if self.lifespan_function is not None:
                self._lifespan_context = self.lifespan_function()
                await self._lifespan_context.__aenter__()

            await self._run_tasks(self.startup_tasks)

    async def _shutdown(self, *suppress_args: typing.Any) -> None:
        if self._started:
            await logger.adebug("Running lifespan shutdown tasks")
            self._started = False

            if self._lifespan_context is not None:
                await self._lifespan_context.__aexit__(*suppress_args)
                self._lifespan_context = None

            await self._run_tasks(self.shutdown_tasks)

    def start(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        run_task(self._start(), loop=loop)

    def shutdown(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        run_task(self._shutdown(None, None, None), loop=loop)

    def on_startup[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.startup_tasks.append(to_coroutine_task(task))
        return task

    def on_shutdown[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.shutdown_tasks.append(to_coroutine_task(task))
        return task


__all__ = ("CoroutineTask", "DelayedTask", "Lifespan", "run_task", "to_coroutine_task")



import dataclasses
import typing
from abc import ABC, abstractmethod
from functools import cached_property

from kungfu.library.monad.result import Error
from nodnod.error import NodeError

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.return_manager.utils import _get_types
from telegrinder.modules import logger
from telegrinder.node.compose import compose
from telegrinder.tools import fullname
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

type ManagerFunction = typing.Callable[..., typing.Any | typing.Awaitable[typing.Any]]


def register_manager(return_type: typing.Any, /) -> typing.Callable[[ManagerFunction], Manager]:
    def wrapper(function: ManagerFunction, /) -> Manager:
        function = function.__func__ if isinstance(function, classmethod | staticmethod) else function
        types = _get_types(return_type)
        return Manager((types,) if not isinstance(types, tuple) else types, function)

    return wrapper


@dataclasses.dataclass
class Manager:
    types: tuple[typing.Any, ...]
    function: ManagerFunction
    agent_cls: type[Agent] | None = None

    async def __call__(self, response: typing.Any, context: Context) -> None:
        ctx = context.copy()
        ctx.handler_response = response

        async with compose(
            self.function,
            ctx,
            agent_cls=self.agent_cls,
        ) as result:
            match result:
                case Error(error):
                    await logger.adebug(
                        "Return manager `{}` failed with error:{}",
                        fullname(self.function),
                        NodeError(f"failed to compose return manager `{fullname(self.function)}`", from_error=error),
                    )


class ABCReturnManager(ABC):
    @property
    @abstractmethod
    def managers(self) -> list[Manager]:
        pass

    @abstractmethod
    async def run(self, response: typing.Any, api: API, update: Update, context: Context) -> None:
        pass


class BaseReturnManager(ABCReturnManager):
    def __repr__(self) -> str:
        return "<{}: {}>".format(fullname(self), self.managers)

    @cached_property
    def managers(self) -> list[Manager]:
        return [
            manager for manager in (vars(BaseReturnManager) | vars(type(self))).values() if isinstance(manager, Manager)
        ]

    async def run(
        self,
        response: typing.Any,
        api: API,
        update: Update,
        context: Context,
    ) -> None:
        for manager in self.managers:
            if typing.Any in manager.types or type(response) in manager.types:
                await logger.adebug(
                    "Running manager `{}` for response of type `{}`",
                    fullname(manager.function),
                    fullname(response),
                )
                await manager(response, context)

    def register_manager(self, return_type: typing.Any, /) -> typing.Callable[[ManagerFunction], Manager]:
        def wrapper(function: ManagerFunction, /) -> Manager:
            manager = register_manager(return_type)(function)
            self.managers.append(manager)
            return manager

        return wrapper


__all__ = (
    "ABCReturnManager",
    "BaseReturnManager",
    "Manager",
    "register_manager",
)



from telegrinder.api.api import API
from telegrinder.api.error import APIError, APIServerError, InvalidTokenError
from telegrinder.api.response import APIResponse
from telegrinder.api.token import Token
from telegrinder.api.validators import validate_token

__all__ = (
    "API",
    "APIError",
    "APIResponse",
    "APIServerError",
    "InvalidTokenError",
    "Token",
    "validate_token",
)



import base64
import typing

import kungfu

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.is_from import IsPrivate
from telegrinder.bot.rules.markup import Markup
from telegrinder.node.nodes.me import BotUsername
from telegrinder.node.nodes.message_entities import MessageEntities
from telegrinder.types.enums import MessageEntityType


class StartCommand(
    ABCRule,
    requires=[
        IsPrivate(),
        Markup(
            [
                "/start <param>",
                "/start",
                "tg://resolve?domain=<bot_username>&start=<param>",
            ]
        ),
    ],
):
    def __init__(
        self,
        validator: typing.Callable[[str], typing.Any | None] | None = None,
        *,
        deep_link: bool | None = None,
        decode_deep_link_param: bool = False,
        param_required: bool = False,
        alias: str | None = None,
    ) -> None:
        self.param_required = param_required
        self.validator = validator
        self.alias = alias
        self.deep_link = deep_link
        self.decode_deep_link_param = decode_deep_link_param

    def check(
        self,
        bot_username: BotUsername,
        message_entities: kungfu.Option[MessageEntities],
        ctx: Context,
    ) -> bool:
        if self.deep_link is not None and all(
            (
                message_entities.map(
                    lambda entities: entities and entities[0].type == MessageEntityType.BOT_COMMAND,
                ).unwrap_or(False),
                self.deep_link and bot_username == ctx.get("bot_username"),
            ),
        ):
            return False

        param: str | None = ctx.pop("param", None)

        if param is not None and self.decode_deep_link_param:
            param = base64.urlsafe_b64decode(param.encode()).decode()

        validated_param = self.validator(param) if self.validator and param is not None else param

        if self.param_required and validated_param is None:
            return False

        ctx.set(self.alias or "param", validated_param)
        return True


__all__ = ("StartCommand",)



from __future__ import annotations

import typing
from reprlib import recursive_repr

from kungfu.library.monad.option import NOTHING, Option, Some
from nodnod.interface.node_from_function import Externals

if typing.TYPE_CHECKING:
    from nodnod.scope import Scope

    from telegrinder.api.api import API
    from telegrinder.bot.cute_types.update import UpdateCute
    from telegrinder.bot.dispatch.router.base import Router
    from telegrinder.types.objects import Update

type Key = str
type AnyValue = typing.Any


class Context(Externals):
    """Low level per event context storage."""

    SELF_CONTEXT_KEYS: typing.Final = frozenset(("context", "ctx"))

    api: API
    update: Update
    raw_update: Update
    update_cute: UpdateCute
    per_event_scope: Scope
    exceptions_update: dict[Router, Exception] = {}
    exception_update: Option[Exception] = NOTHING

    @typing.overload
    def __init__(self) -> None: ...

    @typing.overload
    def __init__(self, map: typing.Mapping[Key, AnyValue], /) -> None: ...

    @typing.overload
    def __init__(self, **kwargs: AnyValue) -> None: ...

    def __init__(
        self,
        map: typing.Mapping[Key, AnyValue] | None = None,
        **kwargs: AnyValue,
    ) -> None:
        Externals.__init__(self, map or kwargs)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(f"{k}={repr(v) if v is not self else '<self>'}" for k, v in self.items()),
        )

    def __setitem__(self, __key: Key, __value: AnyValue) -> None:
        Externals.__setitem__(self, __key, __value)

    def __getitem__(self, __key: Key) -> AnyValue:
        if __key in type(self).SELF_CONTEXT_KEYS:
            return self
        return Externals.__getitem__(self, __key)

    def __delitem__(self, __key: Key) -> None:
        Externals.__delitem__(self, __key)

    def __setattr__(self, __name: str, __value: AnyValue) -> None:
        self.__setitem__(__name, __value)

    def __getattribute__(self, __name: str) -> AnyValue:
        if __name in type(self).SELF_CONTEXT_KEYS:
            return self

        if __name in _CONTEXT_CLASS_ATTRS and not Externals.__contains__(self, __name):
            return super().__getattribute__(__name)

        return self[__name]

    def __delattr__(self, __name: str) -> None:
        self.__delitem__(__name)

    def __contains__(self, __key: object) -> bool:
        if __key in type(self).SELF_CONTEXT_KEYS:
            return True
        return Externals.__contains__(self, __key)

    def __or__(self, other: object, /) -> typing.Self:
        if type(other) is not Context and not isinstance(other, dict):
            return NotImplemented

        new_context = type(self)(self)
        new_context |= other
        return new_context

    def __ior__(self, other: object, /) -> typing.Self:
        if type(other) is not Context and not isinstance(other, dict):
            raise TypeError(f"Cannot update `Context` with `{type(other).__name__}`.")

        for key, value in other.items():
            self[key] = value

        return self

    def as_dict(self) -> dict[Key, AnyValue]:
        return {key: value for key, value in Externals.items(self)}

    def add_roots(
        self,
        api: API,
        update: Update,
        per_event_scope: Scope,
        /,
    ) -> typing.Self:
        from telegrinder.bot.cute_types.update import UpdateCute

        for key, value in {
            "api": api,
            "raw_update": update,
            "update": update,
            "update_cute": UpdateCute.from_update(update, bound_api=api),
            "per_event_scope": per_event_scope,
        }.items():
            self[key] = value

        return self

    def add_exception_update(self, exception_update: Exception, /) -> typing.Self:
        self.exception_update = Some(exception_update)
        return self

    def copy(self) -> typing.Self:
        return type(self)(self)

    def set(self, key: Key, value: AnyValue) -> None:
        self[key] = value

    @typing.overload
    def get(self, key: Key) -> AnyValue | None: ...

    @typing.overload
    def get[T](self, key: Key, default: T) -> T | AnyValue: ...

    @typing.overload
    def get(self, key: Key, default: None = None) -> AnyValue | None: ...

    def get[T](self, key: Key, default: T | None = None) -> T | AnyValue | None:
        return dict.get(self, key, default)

    def get_or_set[T](self, key: Key, default: T) -> T:
        if key not in self:
            self.set(key, default)
        return self.get(key, default)

    def delete(self, key: Key) -> None:
        del self[key]


_CONTEXT_CLASS_ATTRS = frozenset(Context.__dict__ | dict.__dict__ | object.__dict__)


__all__ = ("Context",)



import typing

if typing.TYPE_CHECKING:
    from kungfu.library.monad.option import Option
else:
    import kungfu.library
    import msgspec

    class OptionMeta(type):
        def __instancecheck__(cls, __instance):
            return isinstance(__instance, (kungfu.library.Some | kungfu.library.Nothing, msgspec.UnsetType))

    class Option[Value](metaclass=OptionMeta):
        pass


__all__ = ("Option",)



from telegrinder.model import From, Model, field
from telegrinder.msgspec_utils.custom_types import Literal, Option, datetime

type ChatType = Literal["sender", "private", "group", "supergroup", "channel"]


class WebAppChat(Model):
    """Object `WebAppChat`, see the [documentation](https://core.telegram.org/bots/webapps#webappchat).

    This object contains the data of the Mini App chat.
    """

    id: int = field()
    """Unique identifier for this chat. This number may have more than 32
    significant bits and some programming languages may have
    difficulty/silent defects in interpreting it. But it has at most 52
    significant bits, so a signed 64-bit integer or double-precision float
    type are safe for storing this identifier."""

    type: Literal["group", "supergroup", "channel"] = field()
    """Type of chat, can be either `group`, `supergroup` or `channel`."""

    title: str = field()
    """Title of the chat."""

    username: Option[str] = field(default=..., converter=From[str | None])
    """Optional. Username of the chat."""

    photo_url: Option[str] = field(default=..., converter=From[str | None])
    """Optional. URL of the chat’s photo. The photo can be in .jpeg or .svg formats.
    Only returned for Web Apps launched from the attachment menu."""


class WebAppUser(Model):
    """Object `WebAppUser`, see the [documentation](https://core.telegram.org/bots/webapps#webappuser).

    This object contains the data of the Mini App user.
    """

    id: int = field()
    """A unique identifier for the user or bot. This number may have more than 32
    significant bits and some programming languages may have difficulty/silent
    defects in interpreting it. It has at most 52 significant bits, so a 64-bit
    integer or a double-precision float type is safe for storing this identifier."""

    first_name: str = field()
    """First name of the user or bot."""

    is_bot: Option[bool] = field(default=..., converter=From[bool | None])
    """Optional. True, if this user is a bot. Returns in the receiver field only."""

    last_name: Option[str] = field(default=..., converter=From[str | None])
    """Optional. Last name of the user or bot."""

    username: Option[str] = field(default=..., converter=From[str | None])
    """Optional. Username of the user or bot."""

    language_code: Option[str] = field(default=..., converter=From[str | None])
    """Optional. IETF language tag of the user's language. Returns in user field only."""

    is_premium: Option[bool] = field(default=..., converter=From[bool | None])
    """Optional. True, if this user is a Telegram Premium user."""

    added_to_attachment_menu: Option[bool] = field(default=..., converter=From[bool | None])
    """Optional. True, if this user added the bot to the attachment menu."""

    allows_write_to_pm: Option[bool] = field(default=..., converter=From[bool | None])
    """Optional. True, if this user allowed the bot to message them."""

    photo_url: Option[str] = field(default=..., converter=From[str | None])
    """Optional. URL of the user’s profile photo. The photo can be in .jpeg or .svg formats."""


class WebAppInitData(Model):
    """Object `WebAppInitData`, see the [documentation](https://core.telegram.org/bots/webapps#webappinitdata).

    This object contains data that is transferred to the Mini App when it is opened.
    It is empty if the Mini App was launched from a keyboard button or from inline mode.
    """

    auth_date: datetime = field(converter=From[datetime | int])
    """Unix time when the form was opened."""

    hash: str = field()
    """A hash of all passed parameters, which the bot server can use to check their validity."""

    signature: str = field()
    """A signature of all passed parameters (except hash), which the third party
    can use to check their validity."""

    query_id: Option[str] = field(default=..., converter=From[str | None])
    """Optional. A unique identifier for the Mini App session, required for
    sending messages via the answerWebAppQuery method."""

    user: Option[WebAppUser] = field(default=..., converter=From[WebAppUser | None])
    """Optional. An object containing data about the current user."""

    receiver: Option[WebAppUser] = field(default=..., converter=From[WebAppUser | None])
    """Optional. An object containing data about the chat partner of the current
    user in the chat where the bot was launched via the attachment menu. Returned
    only for private chats and only for Mini Apps launched via the attachment menu."""

    chat: Option[WebAppChat] = field(default=..., converter=From[WebAppChat | None])
    """Optional. An object containing data about the chat where the bot was launched
    via the attachment menu. Returned for supergroups, channels and group chats – only
    for Mini Apps launched via the attachment menu."""

    chat_type: Option[ChatType] = field(default=..., converter=From[ChatType | None])
    """Optional. Type of the chat from which the Mini App was opened. Can be either
    `sender` for a private chat with the user opening the link, `private`, `group`,
    `supergroup`, or `channel`. Returned only for Mini Apps launched from direct links."""

    chat_instance: Option[str] = field(default=..., converter=From[str | None])
    """Optional. Global identifier, uniquely corresponding to the chat from which the
    Mini App was opened. Returned only for Mini Apps launched from a direct link."""

    start_param: Option[str] = field(default=..., converter=From[str | None])
    """Optional. The value of the startattach parameter, passed via link. Only
    returned for Mini Apps when launched from the attachment menu via link.

    The value of the `start_param` parameter will also be passed in the GET-parameter
    `tgWebAppStartParam`, so the Mini App can load the correct interface right away.
    """

    can_send_after: Option[int] = field(default=..., converter=From[int | None])
    """Optional. Time in seconds, after which a message can be sent via the answerWebAppQuery method."""


__all__ = ("WebAppChat", "WebAppInitData", "WebAppUser")



import dataclasses
import typing

from kungfu.library.monad.result import Error, Ok
from nodnod.error import NodeError
from nodnod.interface.generic import generic_node
from nodnod.interface.polymorphic import case, polymorphic
from nodnod.interface.scalar import scalar_node

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.bot.cute_types.shipping_query import ShippingQueryCute
from telegrinder.node.nodes.global_node import GlobalNode
from telegrinder.tools.serialization.abc import ABCDataSerializer
from telegrinder.tools.serialization.json_ser import JSONSerializer

type _UndefinedSerializer = typing.Any


@scalar_node
@polymorphic[str]
class Payload:
    @case
    def compose_callback_query(cls, event: CallbackQueryCute) -> str:
        return event.data.expect(NodeError("CallbackQuery has no data."))

    @case
    def compose_pre_checkout_query(cls, event: PreCheckoutQueryCute) -> str:
        return event.invoice_payload

    @case
    def compose_shipping_query(cls, event: ShippingQueryCute) -> str:
        return event.invoice_payload

    @case
    def compose_message(cls, event: MessageCute) -> str:
        return event.successful_payment.map(
            lambda payment: payment.invoice_payload,
        ).expect(NodeError("Message has no successful payment."))


@dataclasses.dataclass(frozen=True)
class PayloadSerializer[T: type[ABCDataSerializer] = typing.Any](GlobalNode[T]):
    serializer: type[ABCDataSerializer[typing.Any]]

    @classmethod
    def __compose__(cls) -> typing.Self:
        return cls(serializer=JSONSerializer)


@generic_node
class _PayloadData[Data, Serializer: ABCDataSerializer = _UndefinedSerializer]:
    @classmethod
    def __compose__(
        cls,
        payload: Payload,
        data: type[Data],
        payload_serializer: type[Serializer],
        global_serializer: PayloadSerializer,
    ) -> typing.Any:
        if payload_serializer is _UndefinedSerializer:
            serializer = getattr(data, "__serializer__", global_serializer.serializer)
        else:
            serializer = payload_serializer

        match serializer(data).deserialize(payload):
            case Ok(value):
                return value
            case Error(err):
                raise NodeError(err)


if typing.TYPE_CHECKING:
    type PayloadData[
        DataType = typing.Any,
        Serializer: ABCDataSerializer = _UndefinedSerializer,
    ] = typing.Annotated[DataType, Serializer]

else:
    PayloadData = _PayloadData


__all__ = ("Payload", "PayloadData", "PayloadSerializer")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.rules import StartCommand

api = API(token=Token.from_env())
bot = Telegrinder(api)


@bot.on.message(StartCommand(lambda x: (int(x) if x.isdigit() else None)))
async def start_handler(message: Message, param: int | None) -> None:
    if param is None:
        await message.answer("You have no integer start query((")
        return
    await message.answer("Ahah you integer start query is so funny, its {0} and {0}-42={1}".format(param, param - 42))


bot.run_forever()



"""Built-in nodes."""

from telegrinder.node.nodes.attachment import (
    Animation,
    Attachment,
    Audio,
    Document,
    MediaGroup,
    Photo,
    Poll,
    SuccessfulPayment,
    Video,
    VideoNote,
    Voice,
)
from telegrinder.node.nodes.callback_query import CallbackQueryData, CallbackQueryDataJson
from telegrinder.node.nodes.channel import (
    Channel,
    ChannelId,
    ChannelPost,
    ChannelPostId,
    ChatMessageChannelPost,
    ChatMessageChannelPostAuthor,
    ChatMessageChannelPostChannel,
    ChatMessageChannelPostChannelId,
    ChatMessageChannelPostId,
)
from telegrinder.node.nodes.command import CommandInfo
from telegrinder.node.nodes.error import Error
from telegrinder.node.nodes.event import EventNode
from telegrinder.node.nodes.file import File, FileId
from telegrinder.node.nodes.global_node import GlobalNode
from telegrinder.node.nodes.i18n import ABCTranslator, BaseTranslator, I18NConfig, KeySeparator
from telegrinder.node.nodes.me import BotUsername, Me
from telegrinder.node.nodes.message_entities import MessageEntities
from telegrinder.node.nodes.payload import Payload, PayloadData, PayloadSerializer
from telegrinder.node.nodes.reply_message import ReplyMessage
from telegrinder.node.nodes.source import ChatId, ChatSource, Locale, Source, UserId, UserSource
from telegrinder.node.nodes.state_mutator import State, StateMutator
from telegrinder.node.nodes.text import Caption, HTMLCaption, HTMLText, Text, TextInteger, TextLiteral

__all__ = (
    "ABCTranslator",
    "Animation",
    "Attachment",
    "Audio",
    "BaseTranslator",
    "BotUsername",
    "CallbackQueryData",
    "CallbackQueryDataJson",
    "Caption",
    "Channel",
    "ChannelId",
    "ChannelPost",
    "ChannelPostId",
    "ChatId",
    "ChatMessageChannelPost",
    "ChatMessageChannelPostAuthor",
    "ChatMessageChannelPostChannel",
    "ChatMessageChannelPostChannelId",
    "ChatMessageChannelPostId",
    "ChatSource",
    "CommandInfo",
    "Document",
    "Error",
    "EventNode",
    "File",
    "FileId",
    "GlobalNode",
    "HTMLCaption",
    "HTMLText",
    "I18NConfig",
    "KeySeparator",
    "Locale",
    "Me",
    "MediaGroup",
    "MessageEntities",
    "Payload",
    "PayloadData",
    "PayloadSerializer",
    "Photo",
    "Poll",
    "ReplyMessage",
    "Source",
    "State",
    "StateMutator",
    "SuccessfulPayment",
    "Text",
    "TextInteger",
    "TextLiteral",
    "UserId",
    "UserSource",
    "Video",
    "VideoNote",
    "Voice",
)



import enum
import math
import sys
import typing

from telegrinder.modules import logger

NOT_SUPPORTED: typing.Final = "NOT_SUPPORTED"
ENUM_FRIENDS: typing.Final = (str, int, float)
NOT_SUPPORTED_VALUES: typing.Final = {
    str: NOT_SUPPORTED,
    int: math.inf,
    float: sys.maxsize,
}


def _is_friend(bases: tuple[type[typing.Any], ...], /) -> bool:
    return any(friend in bases for friend in ENUM_FRIENDS)


class BaseEnumMeta(enum.EnumMeta, type):
    if typing.TYPE_CHECKING:

        class _BaseEnumMeta(enum.Enum):  # noqa
            NOT_SUPPORTED = enum.auto()

        NOT_SUPPORTED: typing.Literal[_BaseEnumMeta.NOT_SUPPORTED]

    else:

        @staticmethod
        def _member_missing(cls, value):
            logger.warning(
                "Unsupported value {!r} for enum of type {}. Probably telegrinder needs "
                "to be updated to support the latest version of Telegram Bot API.",
                value,
                cls,
            )
            return cls._member_map_["NOT_SUPPORTED"]

        def __new__(
            metacls,
            cls,
            bases,
            classdict,
            *,
            boundary=None,
            _simple=False,
            **kwds,
        ):
            if _is_friend(bases):
                classdict["NOT_SUPPORTED"] = next(
                    (value for base, value in NOT_SUPPORTED_VALUES.items() if base in bases),
                    NOT_SUPPORTED,
                )

            classdict["_missing_"] = classmethod(BaseEnumMeta._member_missing)
            return super().__new__(metacls, cls, bases, classdict, boundary=boundary, _simple=_simple, **kwds)


__all__ = ("BaseEnumMeta",)



import dataclasses
import typing

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.abc import DictStrAny, RawKeyboard


class KeyboardParams(typing.TypedDict, total=False):
    is_persistent: bool
    one_time_keyboard: bool
    resize_keyboard: bool
    is_selective: bool
    input_field_placeholder: str | None


@dataclasses.dataclass(frozen=True)
class KeyboardModel:
    keyboard: RawKeyboard
    is_persistent: bool = dataclasses.field(default=False, kw_only=True)
    one_time_keyboard: bool = dataclasses.field(default=False, kw_only=True)
    resize_keyboard: bool = dataclasses.field(default=False, kw_only=True)
    is_selective: bool = dataclasses.field(default=False, kw_only=True)
    input_field_placeholder: str | None = dataclasses.field(default=None, kw_only=True)

    def dict(self) -> DictStrAny:
        dct = dataclasses.asdict(self)
        dct["keyboard"] = [row for row in self.keyboard if row]
        return dct


__all__ = ("KeyboardModel", "KeyboardParams")



import dataclasses
import enum
import typing

from telegrinder import (
    API,
    CallbackQuery,
    Message,
    Telegrinder,
    Token,
)
from telegrinder.modules import setup_logger
from telegrinder.node import ChatId
from telegrinder.rules import PayloadModelRule, Text
from telegrinder.tools import MsgPackSerializer
from telegrinder.tools.keyboard import InlineButton, InlineKeyboard

setup_logger()

api = API(token=Token.from_env())
bot = Telegrinder(api)


@enum.unique
class MenuAction(enum.Enum):
    show_fact = enum.auto()
    show_store = enum.auto()


@enum.unique
class StoreAction(enum.Enum):
    buy_coffee = enum.auto()
    buy_tea = enum.auto()
    back_to_menu = enum.auto()


@dataclasses.dataclass(slots=True, frozen=True)
class MenuCallback:
    __key__ = "menu"
    __serializer__ = MsgPackSerializer[typing.Self]

    action: MenuAction


@dataclasses.dataclass(slots=True, frozen=True)
class StoreCallback:
    __key__ = "store"
    __serializer__ = MsgPackSerializer[typing.Self]

    action: StoreAction
    item: str = ""
    price: int = 0


class MainMenuKeyboard(InlineKeyboard):
    show_fact = InlineButton(
        "🎯 Show Random Fact",
        callback_data=MenuCallback(action=MenuAction.show_fact),
        new_row=True,
    )
    show_store = InlineButton(
        "🛒 Open Store",
        callback_data=MenuCallback(action=MenuAction.show_store),
    )


class StoreKeyboard(InlineKeyboard):
    buy_coffee = InlineButton(
        "☕ Coffee - $3",
        callback_data=StoreCallback(action=StoreAction.buy_coffee, item="Coffee", price=3),
        new_row=True,
    )
    buy_tea = InlineButton(
        "🍵 Tea - $2",
        callback_data=StoreCallback(action=StoreAction.buy_tea, item="Tea", price=2),
        new_row=True,
    )
    back_to_menu = InlineButton(
        "⬅️ Back to Menu",
        callback_data=StoreCallback(action=StoreAction.back_to_menu),
    )


@bot.on.message(Text("/start"))
async def start(message: Message) -> None:
    await message.answer(
        text="hi there. pick an option below:",
        reply_markup=MainMenuKeyboard.get_markup(),
    )


@bot.on.callback_query(StoreKeyboard.back_to_menu)
async def back_menu(cb: CallbackQuery, chat_id: ChatId) -> None:
    if cb.message_cute:
        await cb.message_cute.unwrap().delete()

    await cb.api.send_message(
        chat_id=chat_id,
        text="hi there. pick an option below:",
        reply_markup=MainMenuKeyboard.get_markup(),
    )
    await cb.answer()


@bot.on.callback_query(MainMenuKeyboard.show_fact)
async def interest_fact(cb: CallbackQuery) -> None:
    await cb.answer("brace yourself: he comes again", show_alert=True)


@bot.on.callback_query(MainMenuKeyboard.show_store)
async def store(cb: CallbackQuery) -> None:
    await cb.edit_text("what do u want?", reply_markup=StoreKeyboard.get_markup())
    await cb.answer()


@bot.on.callback_query(PayloadModelRule(StoreCallback, alias="my_model"))
async def view_store(cb: CallbackQuery, chat_id: ChatId, my_model: StoreCallback) -> None:
    await cb.api.send_message(chat_id=chat_id, text=f"your choice - {my_model.item}; price - {my_model.price}")
    await cb.answer()


bot.run_forever(skip_updates=True)



from functools import cached_property

from kungfu.library.monad.option import NOTHING, Some

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.business_connection import BusinessConnectionCute
from telegrinder.bot.cute_types.business_messages_deleted import BusinessMessagesDeletedCute
from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.chat_boost_removed import ChatBoostRemovedCute
from telegrinder.bot.cute_types.chat_boost_updated import ChatBoostUpdatedCute
from telegrinder.bot.cute_types.chat_join_request import ChatJoinRequestCute
from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.cute_types.chosen_inline_result import ChosenInlineResultCute
from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.message_reaction_count_updated import MessageReactionCountUpdatedCute
from telegrinder.bot.cute_types.message_reaction_updated import MessageReactionUpdatedCute
from telegrinder.bot.cute_types.paid_media_purchased import PaidMediaPurchasedCute
from telegrinder.bot.cute_types.poll import PollCute
from telegrinder.bot.cute_types.poll_answer import PollAnswerCute
from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.bot.cute_types.shipping_query import ShippingQueryCute
from telegrinder.model import From, field
from telegrinder.msgspec_utils import Option
from telegrinder.types.objects import *


class UpdateCute(BaseCute[Update], Update, kw_only=True):
    """Cute version of the object [Update](https://core.telegram.org/bots/api#update).

    This object represents a `cute` incoming update.
    At most one of the optional parameters can be present in any given cute update.
    """

    message: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New incoming message of any kind - text, photo, sticker, etc."""

    edited_message: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New version of a message that is known to the bot and was edited.
    This update may at times be triggered by changes to message fields that are
    either unavailable or not actively used by your bot."""

    channel_post: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New incoming channel post of any kind - text, photo, sticker,
    etc."""

    edited_channel_post: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New version of a channel post that is known to the bot and was edited.
    This update may at times be triggered by changes to message fields that are
    either unavailable or not actively used by your bot."""

    business_connection: Option[BusinessConnectionCute] = field(
        default=...,
        converter=From["BusinessConnectionCute | None"],
    )
    """Optional. The bot was connected to or disconnected from a business account,
    or a user edited an existing connection with the bot."""

    business_message: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New message from a connected business account."""

    edited_business_message: Option[MessageCute] = field(
        default=...,
        converter=From[MessageCute | None],
    )
    """Optional. New version of a message from a connected business account."""

    deleted_business_messages: Option[BusinessMessagesDeletedCute] = field(
        default=...,
        converter=From["BusinessMessagesDeletedCute | None"],
    )
    """Optional. Messages were deleted from a connected business account."""

    message_reaction: Option[MessageReactionUpdatedCute] = field(
        default=...,
        converter=From["MessageReactionUpdatedCute | None"],
    )
    """Optional. A reaction to a message was changed by a user. The bot must be an
    administrator in the chat and must explicitly specify `message_reaction`
    in the list of allowed_updates to receive these updates. The update isn't
    received for reactions set by bots."""

    message_reaction_count: Option[MessageReactionCountUpdatedCute] = field(
        default=...,
        converter=From["MessageReactionCountUpdatedCute | None"],
    )
    """Optional. Reactions to a message with anonymous reactions were changed.
    The bot must be an administrator in the chat and must explicitly specify
    `message_reaction_count` in the list of allowed_updates to receive these
    updates. The updates are grouped and can be sent with delay up to a few minutes."""

    inline_query: Option[InlineQueryCute] = field(
        default=...,
        converter=From[InlineQueryCute | None],
    )
    """Optional. New incoming inline query."""

    chosen_inline_result: Option[ChosenInlineResultCute] = field(
        default=...,
        converter=From["ChosenInlineResultCute | None"],
    )
    """Optional. The result of an inline query that was chosen by a user and sent
    to their chat partner. Please see our documentation on the feedback collecting
    for details on how to enable these updates for your bot."""

    callback_query: Option[CallbackQueryCute] = field(
        default=...,
        converter=From[CallbackQueryCute | None],
    )
    """Optional. New incoming callback query."""

    shipping_query: Option[ShippingQueryCute] = field(
        default=...,
        converter=From[ShippingQueryCute | None],
    )
    """Optional. New incoming shipping query. Only for invoices with flexible
    price."""

    pre_checkout_query: Option[PreCheckoutQueryCute] = field(
        default=...,
        converter=From[PreCheckoutQueryCute | None],
    )
    """Optional. New incoming pre-checkout query. Contains full information
    about checkout."""

    purchased_paid_media: Option[PaidMediaPurchasedCute] = field(
        default=...,
        converter=From["PaidMediaPurchasedCute | None"],
    )
    """Optional. A user purchased paid media with a non-empty payload sent by the
    bot in a non-channel chat."""

    poll: Option[PollCute] = field(
        default=...,
        converter=From["PollCute | None"],
    )
    """Optional. New poll state. Bots receive only updates about manually stopped
    polls and polls, which are sent by the bot."""

    poll_answer: Option[PollAnswerCute] = field(
        default=...,
        converter=From["PollAnswerCute | None"],
    )
    """Optional. A user changed their answer in a non-anonymous poll. Bots receive
    new votes only in polls that were sent by the bot itself."""

    my_chat_member: Option[ChatMemberUpdatedCute] = field(
        default=...,
        converter=From[ChatMemberUpdatedCute | None],
    )
    """Optional. The bot's chat member status was updated in a chat. For private
    chats, this update is received only when the bot is blocked or unblocked
    by the user."""

    chat_member: Option[ChatMemberUpdatedCute] = field(
        default=...,
        converter=From[ChatMemberUpdatedCute | None],
    )
    """Optional. A chat member's status was updated in a chat. The bot must be an
    administrator in the chat and must explicitly specify `chat_member` in
    the list of allowed_updates to receive these updates."""

    chat_join_request: Option[ChatJoinRequestCute] = field(
        default=...,
        converter=From[ChatJoinRequestCute | None],
    )
    """Optional. A request to join the chat has been sent. The bot must have the can_invite_users
    administrator right in the chat to receive these updates."""

    chat_boost: Option[ChatBoostUpdatedCute] = field(
        default=...,
        converter=From["ChatBoostUpdatedCute | None"],
    )
    """Optional. A chat boost was added or changed. The bot must be an administrator
    in the chat to receive these updates."""

    removed_chat_boost: Option[ChatBoostRemovedCute] = field(
        default=...,
        converter=From["ChatBoostRemovedCute | None"],
    )
    """Optional. A boost was removed from a chat. The bot must be an administrator
    in the chat to receive these updates."""

    @cached_property
    def incoming_update(self) -> BaseCute:
        return getattr(self, self.update_type.value).unwrap()

    def get_event[T: BaseCute](self, event_model: type[T], /) -> Option[T]:
        if isinstance(self.incoming_update, event_model):
            return Some(self.incoming_update)
        return NOTHING


__all__ = ("UpdateCute",)



from telegrinder.bot.dispatch.view.abc import ABCView
from telegrinder.bot.dispatch.view.base import ErrorView, EventModelView, EventView, RawEventView, View
from telegrinder.bot.dispatch.view.box import ViewBox
from telegrinder.bot.dispatch.view.media_group import MediaGroupView

__all__ = (
    "ABCView",
    "ErrorView",
    "EventModelView",
    "EventView",
    "MediaGroupView",
    "RawEventView",
    "View",
    "ViewBox",
)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import ShippingQuery, User


class ShippingQueryCute(BaseCute[ShippingQuery], ShippingQuery, kw_only=True):
    @property
    def from_user(self) -> User:
        return self.from_


__all__ = ("ShippingQueryCute",)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import ChatBoostUpdated


class ChatBoostUpdatedCute(BaseCute[ChatBoostUpdated], ChatBoostUpdated, kw_only=True):
    pass


__all__ = ("ChatBoostUpdatedCute",)



import dataclasses
import typing

from telegrinder.tools.keyboard.abc import ABCKeyboard, DictStrAny, RawKeyboard
from telegrinder.tools.keyboard.abc import Button as ButtonType
from telegrinder.tools.keyboard.base import BaseKeyboard
from telegrinder.tools.keyboard.button import BaseButton, Button, InlineButton
from telegrinder.tools.keyboard.data import KeyboardModel, KeyboardParams
from telegrinder.tools.keyboard.utils import (
    RowButtons,
    bound_keyboard_method,
    copy_keyboard,
    init_keyboard,
)
from telegrinder.types.objects import InlineKeyboardMarkup, ReplyKeyboardMarkup, ReplyKeyboardRemove

if typing.TYPE_CHECKING:
    from telegrinder.tools.serialization.abc import ABCDataSerializer


class Keyboard(BaseKeyboard[Button], ABCKeyboard):
    __button_class__ = Button
    __slots__ = ("keyboard_model",)

    def __init_subclass__(cls, *, max_in_row: int = 0, **kwargs: typing.Unpack[KeyboardParams]) -> None:
        cls.__keyboard_instance__ = init_keyboard(cls(**kwargs), max_in_row=max_in_row)

    def __init__(
        self,
        *,
        keyboard_model: KeyboardModel | None = None,
        **kwargs: typing.Unpack[KeyboardParams],
    ) -> None:
        self.keyboard_model = keyboard_model or KeyboardModel(keyboard=[[]], **kwargs)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: keyboard_model={self.keyboard_model!r}>"

    @property
    def keyboard(self) -> RawKeyboard:
        return self.keyboard_model.keyboard

    @property
    def is_persistent(self) -> bool:
        return self.keyboard_model.is_persistent

    @property
    def one_time_keyboard(self) -> bool:
        return self.keyboard_model.one_time_keyboard

    @property
    def resize_keyboard(self) -> bool:
        return self.keyboard_model.resize_keyboard

    @property
    def is_selective(self) -> bool:
        return self.keyboard_model.is_selective

    @property
    def input_field_placeholder(self) -> str | None:
        return self.keyboard_model.input_field_placeholder

    @bound_keyboard_method
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        keyboard_model = dataclasses.replace(
            self.keyboard_model,
            keyboard=copy_keyboard(self.keyboard),
            **with_changes,
        )
        return type(self)(keyboard_model=keyboard_model)

    @bound_keyboard_method
    def dict(self) -> DictStrAny:
        return self.keyboard_model.dict()

    @bound_keyboard_method
    def get_markup(self) -> ReplyKeyboardMarkup:
        return ReplyKeyboardMarkup.from_dict(self.dict())

    @bound_keyboard_method
    def get_keyboard_remove(self) -> ReplyKeyboardRemove:
        return ReplyKeyboardRemove(remove_keyboard=True, selective=self.is_selective)

    def placeholder(self, value: str | None, /) -> typing.Self:
        return self.copy(input_field_placeholder=value)

    def resize(self) -> typing.Self:
        return self.copy(resize_keyboard=True)

    def one_time(self) -> typing.Self:
        return self.copy(one_time_keyboard=True)

    def selective(self) -> typing.Self:
        return self.copy(is_selective=True)

    def persistent(self) -> typing.Self:
        return self.copy(is_persistent=True)

    def no_resize(self) -> typing.Self:
        return self.copy(resize_keyboard=False)

    def no_one_time(self) -> typing.Self:
        return self.copy(one_time_keyboard=False)

    def no_selective(self) -> typing.Self:
        return self.copy(is_selective=False)

    def no_persistent(self) -> typing.Self:
        return self.copy(is_persistent=False)


class InlineKeyboard(BaseKeyboard[InlineButton], ABCKeyboard):
    __button_class__ = InlineButton
    __serializer__: ABCDataSerializer | None = None
    __slots__ = ("keyboard", "serializer")

    def __init_subclass__(
        cls,
        *,
        max_in_row: int = 0,
        serializer: ABCDataSerializer | None = None,
    ) -> None:
        cls.__keyboard_instance__ = init_keyboard(cls(), max_in_row=max_in_row)
        cls.__serializer__ = serializer

    def __init__(
        self,
        *,
        serializer: ABCDataSerializer | None = None,
    ) -> None:
        self.keyboard = [[]]
        self.serializer = serializer or self.__serializer__

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: keyboard={self.keyboard!r}>"

    def add(self, button: ButtonType, /) -> typing.Self:
        if isinstance(button, BaseButton | RowButtons):
            for b in button.buttons if isinstance(button, RowButtons) else (button,):
                if isinstance(b, InlineButton) and b.callback_data_serializer is None:
                    b.callback_data_serializer = self.serializer

        return super().add(button)

    @bound_keyboard_method
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        new_keyboard = type(self)()
        new_keyboard.keyboard = copy_keyboard(self.keyboard)
        return new_keyboard

    @bound_keyboard_method
    def dict(self) -> DictStrAny:
        return dict(inline_keyboard=[row for row in self.keyboard if row])

    @bound_keyboard_method
    def get_markup(self) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup.from_dict(self.dict())


__all__ = ("InlineKeyboard", "Keyboard")



import keyword
import types
import typing
from functools import cache
from reprlib import recursive_repr

import msgspec
from kungfu.library.monad.option import NOTHING

from telegrinder.msgspec_utils import Option, decoder, encoder, is_none, struct_asdict

type From[T] = T

UNSET: typing.Final = typing.cast("typing.Any", msgspec.UNSET)
MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "dict": True,
    "rename": {kw + "_": kw for kw in keyword.kwlist},
}


def field(**kwargs: typing.Any) -> typing.Any:
    if kwargs.get("default") is Ellipsis:
        kwargs["default"] = UNSET

    kwargs.pop("converter", None)
    return msgspec.field(**kwargs)


class Model(msgspec.Struct, **MODEL_CONFIG):
    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        from telegrinder.tools.member_descriptor_proxy import MemberDescriptorProxy

        result = super().__init_subclass__(*args, **kwargs)

        for field_name in getattr(cls, "__slots__", ()):
            setattr(cls, field_name, MemberDescriptorProxy(getattr(cls, field_name)))

        return result

    def __getattribute__(self, name: str, /) -> typing.Any:
        class_ = type(self)
        val = object.__getattribute__(self, name)

        if name not in class_.__struct_fields__:
            return val

        if name in class_.get_optional_fields():
            return NOTHING if val is UNSET else val

        if val is UNSET:
            raise AttributeError(f"{class_.__name__!r} object has no attribute {name!r}")

        return val

    def __post_init__(self) -> None:
        for field, value in struct_asdict(self, exclude_unset=False).items():
            if is_none(value):
                setattr(self, field, UNSET)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(
                f"{f}={val!r}" for f, val in struct_asdict(self, exclude_unset=False, unset_as_nothing=True).items()
            ),
        )

    @classmethod
    @cache
    def get_fields(cls) -> types.MappingProxyType[str, msgspec.inspect.Field]:
        return types.MappingProxyType(
            mapping={f.name: f for f in msgspec.inspect.type_info(cls).fields},  # type: ignore
        )

    @classmethod
    @cache
    def get_optional_fields(cls) -> frozenset[str]:
        return frozenset(
            f.name
            for f in cls.get_fields().values()
            if isinstance(f.type, msgspec.inspect.CustomType) and issubclass(f.type.cls, Option)  # type: ignore
        )

    @classmethod
    def from_data[**P, T](cls: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return decoder.convert(msgspec.structs.asdict(cls(*args, **kwargs)), type=cls)  # type: ignore

    @classmethod
    def from_dict(cls, obj: dict[str, typing.Any], /) -> typing.Self:
        return decoder.convert(obj, type=cls)

    @classmethod
    def from_raw(cls, raw: str | bytes, /) -> typing.Self:
        return decoder.decode(raw, type=cls)

    def _to_dict(
        self,
        dct_name: str,
        exclude_fields: set[str],
        full: bool,
    ) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            self.__dict__[dct_name] = (  # type: ignore
                struct_asdict(self)
                if not full
                else encoder.to_builtins(self.to_dict(exclude_fields=exclude_fields), order="deterministic")
            )

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

    def to_raw(self) -> str:
        return encoder.encode(self)

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        return self._to_dict("model_as_dict", exclude_fields or set(), full=False)

    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        return self._to_dict("model_as_full_dict", exclude_fields or set(), full=True)


__all__ = ("UNSET", "From", "Model", "field", "is_none")



import dataclasses
import typing
from collections import deque

from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.middleware.filter import FilterMiddleware
from telegrinder.bot.dispatch.middleware.media_group import MediaGroupMiddleware
from telegrinder.tools.singleton import Singleton


@dataclasses.dataclass(frozen=True)
class MiddlewareBox(Singleton):
    filter: FilterMiddleware = dataclasses.field(default_factory=FilterMiddleware)
    media_group: MediaGroupMiddleware = dataclasses.field(default_factory=MediaGroupMiddleware)
    _custom_middlewares: deque[ABCMiddleware] = dataclasses.field(
        default_factory=deque,
        repr=False,
    )

    def __call__[Middleware: ABCMiddleware](self, middleware: type[Middleware], /) -> type[Middleware]:
        self.put(middleware())
        return middleware

    def __iter__(self) -> typing.Iterator[ABCMiddleware]:
        for middleware in (self.filter, self.media_group, *self._custom_middlewares):
            yield middleware

    def put(self, middleware: ABCMiddleware, /) -> None:
        self._custom_middlewares.append(middleware)


__all__ = ("MiddlewareBox",)



import dataclasses
import typing
from collections import deque

from kungfu.library.monad.result import Error, Result
from nodnod.agent.event_loop.agent import EventLoopAgent
from nodnod.error import NodeError

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.modules import logger
from telegrinder.node.compose import compose
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

    from telegrinder.bot.rules.abc import ABCRule

type Function[**P = ..., R = typing.Any] = typing.Callable[P, R]


@dataclasses.dataclass(repr=False, slots=True)
class FuncHandler[T: Function](ABCHandler):
    function: T
    rules: dataclasses.InitVar[typing.Iterable[ABCRule] | None] = dataclasses.field(default=None)
    agent: dataclasses.InitVar[type[Agent] | None] = dataclasses.field(default=None, kw_only=True)
    final: bool = dataclasses.field(default=True, kw_only=True)
    preset_context: Context | None = dataclasses.field(default=None, kw_only=True)

    def __post_init__(self, rules: typing.Iterable[ABCRule] | None, agent: type[Agent] | None) -> None:
        self.check_rules = deque(rules or ())
        self.agent_cls = agent or EventLoopAgent

    def __repr__(self) -> str:
        return fullname(self.function)

    @property
    def __call__(self) -> Function:
        return self.function

    async def run(
        self,
        api: API,
        update: Update,
        context: Context,
        check: bool = True,
    ) -> Result[typing.Any, str]:
        if self.preset_context:
            context |= self.preset_context

        if check and self.check_rules:
            await logger.adebug("Checking rules for handler `{!r}`...", self)

            for rule in self.check_rules:
                if not await check_rule(rule, context):
                    return Error(f"Rule {rule!r} failed.")

            await logger.adebug("Rules passed, composing nodes and running handler `{!r}`...", self)
        else:
            await logger.adebug("Composing nodes and running handler `{!r}`...", self)

        async with compose(self.function, context, agent_cls=self.agent_cls) as result:
            return result.map_err(
                lambda error: "{}\n".format(
                    NodeError(f"* failed to compose handler `{self!r}`", from_error=error),
                ),
            )


__all__ = ("FuncHandler",)



from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node

from telegrinder.bot.cute_types.message import MessageCute


@scalar_node
class ReplyMessage:
    @classmethod
    def __compose__(cls, message: MessageCute) -> MessageCute:
        return message.reply_to_message.expect(NodeError("Message doesn't have reply"))


__all__ = ("ReplyMessage",)



import dataclasses
import datetime

from telegrinder.tools.state_mutator import State, StateMutator, mutation

# State declaration


@dataclasses.dataclass
class AliveState(State):
    __description__ = "living"

    since: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)

    @mutation
    def die(self, reason: str) -> "DeadState":
        return DeadState(reason)


@dataclasses.dataclass
class DeadState(State):
    reason: str

    @mutation
    def resurrect(self) -> "AliveState":
        return AliveState()

    @property
    def __description__(self) -> str:
        return f"not living because you died of {self.reason}"


# primary mutations:

be_born = mutation(AliveState)

# or as function


@mutation
def login_as_ghost(silently: bool = False):
    if not silently:
        print("Ghost just logged in ~*_*~")
    return DeadState(reason="~*being a ghost*~")


# State use


from telegrinder import API, Bot, Token, setup_logger
from telegrinder.rules import Text

bot = Bot(API(Token.from_env()))
setup_logger()


@bot.on.message(Text("/die"))
async def die_handler(alive: AliveState):
    new = await alive.die("sadness")
    return f"You died because of {new.reason}"


@bot.on.message(Text("/resurrect"))
async def resurrect_handler(dead: DeadState):
    await dead.resurrect()
    return "You resurrected"


@bot.on.message(Text("/trick"))
async def trick_handler(mutator: StateMutator):
    # To inject mutation from any state: await AliveState().bind(mutator).die("oops")
    await (await be_born(mutator)).die("oops")
    return "Trick was unsuccessful"


@bot.on.message(Text("/song"))
async def sing_a_song_handler(state: AliveState):
    return (
        f"You sing a song: la-lala so great to be living for {(datetime.datetime.now() - state.since).seconds} seconds"
    )


@bot.on.message(Text("Gh0$T_рa$$w0rd"))
async def ghost_handler(m: StateMutator):
    await login_as_ghost(m, silently=True)
    return "~~~🎑♓️🐪❄️"


@bot.on.message()
async def in_state_handler(state: AliveState | DeadState):
    return f"You are {state.__description__}"


@bot.on.message(Text("/be_born"))
async def be_born_handler(mutator: StateMutator):
    await be_born(mutator)
    return "You were born"


@bot.on.message()
async def not_born_handler():
    return "You are not born"


bot.run_forever()



import base64
import typing
from contextlib import suppress
from datetime import timedelta

import msgspec
from kungfu.library import Nothing, Result, Some, Sum, unwrapping
from kungfu.library.monad.option import NOTHING

from telegrinder.api.api import APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.bot.cute_types.message import (
    MessageCute,
    MessageEditShortcuts,
    ReplyMarkup,
    execute_method_edit,
)
from telegrinder.model import UNSET, From, field
from telegrinder.msgspec_utils import Option, decoder
from telegrinder.types.methods_utils import get_params
from telegrinder.types.objects import *

CACHED_CALLBACK_DATA_KEY: typing.Final = "cached_callback_data"


class CallbackQueryCute(BaseCute[CallbackQuery], MessageEditShortcuts, CallbackQuery, kw_only=True):
    message: Option[Sum[MessageCute, InaccessibleMessage]] = field(
        default=UNSET,
        converter=From[MessageCute | InaccessibleMessage | None],
    )
    """Optional. Message sent by the bot with the callback button that originated
    the query."""

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    @unwrapping
    def message_cute(self) -> Option[MessageCute]:
        return self.message.unwrap().only().cast(Some, Nothing)

    @property
    def chat_id(self) -> Option[int]:
        """Optional. Message from chat ID. This will be present if the message is sent
        by the bot with the callback button that originated the query.
        """
        return self.message.map(lambda m: m.v.chat.id)

    @property
    def is_topic_message(self) -> Option[bool]:
        """Optional. True, if the message is a topic message with a name,
        color and icon. This will be present if the message is sent
        by the bot with the callback button that originated the query.
        """
        return self.message.map(
            lambda m: m.only().map(lambda m: m.is_topic_message.unwrap_or(False)).unwrap_or(False),
        )

    @property
    @unwrapping
    def message_thread_id(self) -> Option[int]:
        """Optional. Unique identifier of the target message thread (for forum supergroups only).
        This will be present if the message is sent
        by the bot with the callback button that originated the query.
        """
        return self.message.unwrap().only().map(lambda m: m.message_thread_id.unwrap()).cast(Some, Nothing)

    @property
    def message_id(self) -> Option[int]:
        """Optional. Unique message identifier inside this chat. This will be present
        if the message is sent by the bot with the callback button that originated the query.
        """
        return self.message.map(lambda m: m.v.message_id)

    @property
    def chat(self) -> Option[Chat]:
        """Optional. Chat the callback query originated from. This will be present
        if the message is sent by the bot with the callback button that originated the query.
        """
        return self.message.map(lambda m: m.v.chat)

    @typing.overload
    def decode_data(self) -> Option[dict[str, typing.Any]]: ...

    @typing.overload
    def decode_data[T](self, *, to: type[T]) -> Option[T]: ...

    def decode_data[T](self, *, to: type[T] = dict[str, typing.Any]) -> Option[T]:
        if not self.data:
            return NOTHING

        keys = typing.cast(
            "dict[type[typing.Any], typing.Any]",
            self.__dict__.setdefault(CACHED_CALLBACK_DATA_KEY, {}),  # type: ignore
        )
        if to in keys:
            return keys[to]

        data = NOTHING
        orig_to = typing.get_origin(to) or to

        with suppress(msgspec.ValidationError, msgspec.DecodeError):
            data = (
                Some(decoder.decode(self.data.unwrap(), type=to))
                if not issubclass(orig_to, str | bytes)
                else self.data
                if issubclass(orig_to, str)
                else Some(base64.urlsafe_b64decode(self.data.unwrap()))
            )

        keys[to] = data
        return data  # type: ignore

    @shortcut("answer_callback_query", custom_params={"callback_query_id"})
    async def answer(
        self,
        text: str | None = None,
        *,
        cache_time: timedelta | int | None = None,
        callback_query_id: str | None = None,
        show_alert: bool | None = None,
        url: str | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_callback_query()`, see the [documentation](https://core.telegram.org/bots/api#answercallbackquery)

        Use this method to send answers to callback queries sent from inline keyboards.
        The answer will be displayed to the user as a notification at the top of the
        chat screen or as an alert. On success, True is returned."""
        params = compose_method_params(get_params(locals()), self, default_params={("callback_query_id", "id")})
        return await self.ctx_api.answer_callback_query(**params)

    @shortcut(
        "copy_message",
        custom_params={
            "message_thread_id",
            "chat_id",
            "message_id",
            "from_chat_id",
            "reply_markup",
        },
    )
    async def copy(
        self,
        chat_id: int | str | None = None,
        *,
        allow_paid_broadcast: bool | None = None,
        caption: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        direct_messages_topic_id: int | None = None,
        disable_notification: bool | None = None,
        from_chat_id: int | str | None = None,
        message_effect_id: str | None = None,
        message_id: int | None = None,
        message_thread_id: str | None = None,
        parse_mode: str | None = None,
        protect_content: bool | None = None,
        reply_markup: ReplyMarkup | None = None,
        reply_parameters: ReplyParameters | None = None,
        show_caption_above_media: bool | None = None,
        suggested_post_parameters: SuggestedPostParameters | None = None,
        video_start_timestamp: timedelta | int | None = None,
        **other: typing.Any,
    ) -> Result[MessageId, APIError]:
        """Shortcut `API.copy_message()`, see the [documentation](https://core.telegram.org/bots/api#copymessage)

        Use this method to copy messages of any kind. Service messages, paid media
        messages, giveaway messages, giveaway winners messages, and invoice
        messages can't be copied. A quiz poll can be copied only if the value of the
        field correct_option_id is known to the bot. The method is analogous to
        the method forwardMessage, but the copied message doesn't have a link to
        the original message. Returns the MessageId of the sent message on success."""
        return await MessageCute.copy(self, **get_params(locals()))  # type: ignore

    @shortcut("delete_message", custom_params={"message_thread_id", "chat_id", "message_id"})
    async def delete(
        self,
        *,
        chat_id: int | None = None,
        message_id: int | None = None,
        message_thread_id: str | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.delete_message()`, see the [documentation](https://core.telegram.org/bots/api#deletemessage)

        Use this method to delete a message, including service messages, with the
        following limitations: - A message can only be deleted if it was sent less
        than 48 hours ago. - Service messages about a supergroup, channel, or forum
        topic creation can't be deleted. - A dice message in a private chat can only
        be deleted if it was sent more than 24 hours ago. - Bots can delete outgoing
        messages in private chats, groups, and supergroups. - Bots can delete incoming
        messages in private chats. - Bots granted can_post_messages permissions
        can delete outgoing messages in channels. - If the bot is an administrator
        of a group, it can delete any message there. - If the bot has can_delete_messages
        administrator right in a supergroup or a channel, it can delete any message
        there. - If the bot has can_manage_direct_messages administrator right
        in a channel, it can delete any message in the corresponding direct messages
        chat. Returns True on success."""
        return await MessageCute.delete(self, **get_params(locals()))  # type: ignore

    @shortcut(
        "edit_message_text",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_text(
        self,
        text: str,
        *,
        business_connection_id: str | None = None,
        chat_id: int | str | None = None,
        entities: list[MessageEntity] | None = None,
        inline_message_id: str | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        message_id: int | None = None,
        message_thread_id: str | None = None,
        parse_mode: str | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Sum[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_text()`, see the [documentation](https://core.telegram.org/bots/api#editmessagetext)

        Use this method to edit text and game messages. On success, if the edited
        message is not an inline message, the edited Message is returned, otherwise
        True is returned. Note that business messages that were not sent by the bot
        and do not contain an inline keyboard can only be edited within 48 hours from
        the time they were sent.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param text: New text of the message, 1-4096 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the message text. See formatting options formore details.

        :param entities: A JSON-serialized list of special entities that appear in message text,which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""
        ...


__all__ = ("CallbackQueryCute",)



import typing
from functools import cached_property

from telegrinder.api.error import InvalidTokenError
from telegrinder.modules import take_env


class Token(str):
    def __new__(cls, token: str, /) -> typing.Self:
        if token.count(":") != 1 or not token.split(":")[0].isdigit():
            raise InvalidTokenError("Invalid token format, it should look like 12345:ABCdef")
        return super().__new__(cls, token)

    def __repr__(self) -> str:
        return f"<Token: {self.bot_id}:{self.token[:9]}...>"

    @classmethod
    def from_env(cls, var_name: str = "BOT_TOKEN") -> typing.Self:
        return cls(take_env(var_name))

    @cached_property
    def token(self) -> str:
        return self.split(":")[1]

    @cached_property
    def bot_id(self) -> int:
        return int(self.split(":")[0])


__all__ = ("Token",)



from .methods import *
from .objects import *

__all__ = (
    "GeneratorMethods",
    "GeneratorObjects",
    "MethodsParamsAnnotations",
    "MethodsParamsAnnotationsAnnotations",
    "MethodsParamsAnnotationsAnnotationsParam",
    "MethodsParamsLiteralTypes",
    "MethodsParamsLiteralTypesParam",
    "ObjectsFields",
    "ObjectsFieldsAnnotations",
    "ObjectsFieldsAnnotationsAnnotations",
    "ObjectsFieldsAnnotationsAnnotationsField",
    "ObjectsFieldsDefaults",
    "ObjectsFieldsIdByDefault",
    "ObjectsFieldsIdByDefaultField",
    "ObjectsFieldsLiteralTypes",
    "ObjectsFieldsLiteralTypesField",
)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager
from telegrinder.tools.formatting import HTML


class MessageReturnManager(BaseReturnManager):
    @register_manager(str)
    @staticmethod
    async def str_manager(handler_response: str, event: MessageCute) -> None:
        await event.answer(handler_response)

    @register_manager(list | tuple)
    @staticmethod
    async def seq_manager(
        handler_response: list[typing.Any] | tuple[typing.Any, ...],
        event: MessageCute,
    ) -> None:
        for message in handler_response:
            await event.answer(str(message))

    @register_manager(dict)
    @staticmethod
    async def dict_manager(handler_response: dict[str, typing.Any], event: MessageCute) -> None:
        await event.answer(**handler_response)

    @register_manager(HTML)
    @staticmethod
    async def html_manager(handler_response: HTML, event: MessageCute) -> None:
        await event.answer(handler_response, parse_mode=HTML.PARSE_MODE)


__all__ = ("MessageReturnManager",)



import typing

from telegrinder import API, Telegrinder, Token
from telegrinder.modules import setup_logger
from telegrinder.node import EventNode
from telegrinder.rules import IsUpdateType, Text
from telegrinder.types import UpdateType
from telegrinder.types.objects import Message

setup_logger()

bot = Telegrinder(API(Token.from_env()))


@bot.on.raw(Text(["hello", "hi"], ignore_case=True))
async def handle_raw_message(raw_msg: EventNode[Message]) -> None:
    await bot.api.send_message(chat_id=raw_msg.chat_id, text="Hello, World!")


@bot.on.raw(IsUpdateType(UpdateType.CALLBACK_QUERY))
async def handle_raw_cb(event: EventNode[dict[str, typing.Any]]) -> None:
    await bot.api.answer_callback_query(callback_query_id=event["id"])


bot.run_forever(skip_updates=True)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import MessageReactionUpdated


class MessageReactionUpdatedCute(BaseCute[MessageReactionUpdated], MessageReactionUpdated, kw_only=True):
    pass


__all__ = ("MessageReactionUpdatedCute",)



import typing

from kungfu.library.monad.option import Some
from nodnod.error import NodeError
from nodnod.node import Node

from telegrinder.node.nodes.payload import PayloadSerializer
from telegrinder.node.nodes.source import Source
from telegrinder.tools.fullname import fullname
from telegrinder.tools.serialization import ABCDataSerializer
from telegrinder.tools.state_storage.memory import ABCStateStorage, MemoryStateStorage


class StateMutator(Node):
    STORAGE = MemoryStateStorage[str]()  # TODO: use nodnod injection to get storage inside StateMutator.compose
    KEY_MAP: dict[str, type[State]] = {}

    def __init__(
        self,
        storage: ABCStateStorage[str],
        user_id: int,
        serializer: type[ABCDataSerializer[State]],
    ) -> None:
        self.storage = storage
        self.user_id = user_id
        self.serializer = serializer

    async def get(self) -> State | None:
        match await self.storage.get(self.user_id):
            case Some(state_data) if state_data.key in self.KEY_MAP:
                return (
                    self.serializer(self.KEY_MAP[state_data.key])
                    .deserialize(state_data.payload)
                    .map(lambda state: state.bind(self))
                    .unwrap_or_none()
                )

        return None

    async def set(self, state: State) -> None:
        state_cls = state.__class__
        key = fullname(state_cls)
        payload = self.serializer(state_cls).serialize(state)
        await self.storage.set(self.user_id, key, payload)
        self.KEY_MAP[key] = state_cls

    @classmethod
    def __compose__(cls, src: Source, serializer: PayloadSerializer) -> typing.Self:
        return cls(cls.STORAGE, src.from_user.id, serializer.serializer)


class State:
    def bind(self, mutator: StateMutator) -> typing.Self:
        self.__mutator__ = mutator
        return self

    async def enter(self) -> None:
        await self.__mutator__.set(state=self)

    async def exit(self) -> None:
        pass

    @classmethod
    async def __compose__(cls, mutator: StateMutator) -> typing.Self:
        current_state = await mutator.get()
        if current_state is None or not isinstance(current_state, cls):
            raise NodeError("State mismatch.")
        return current_state


__all__ = ("State", "StateMutator")



import hashlib
import hmac
import typing

SECRET_TOKEN_KEY: typing.Final = "X-Telegram-Bot-Api-Secret-Token"


def verify_secret_token(
    secret_token: str,
    request_headers: typing.Mapping[str, typing.Any],
) -> bool:
    """Verifies update request is from telegram."""
    return request_headers.get(SECRET_TOKEN_KEY) == secret_token


def webapp_validate_request(
    bot_token: str,
    request_query_params: typing.Mapping[str, typing.Any],
) -> bool:
    """Verifies authentity of webapp request by counting hash of its parameters."""
    items = sorted(request_query_params.items(), key=lambda kv: kv[0])
    data_check_string = "\n".join(f"{k}={param}" for k, param in items if k != "hash")
    secret = hmac.new(
        "WebAppData".encode(),
        bot_token.encode(),
        hashlib.sha256,
    ).digest()
    data_chk = hmac.new(secret, data_check_string.encode(), hashlib.sha256)

    if (hash_ := request_query_params.get("hash")) is None:
        return False
    return hmac.compare_digest(data_chk.hexdigest(), hash_)


__all__ = ("verify_secret_token", "webapp_validate_request")



import typing

from telegrinder.tools.aio import maybe_awaitable

if typing.TYPE_CHECKING:
    from telegrinder.node import State, StateMutator

    class BoundMethod[**P, T, R](typing.Protocol):
        @staticmethod
        def __call__(self: T, *args: P.args, **kwargs: P.kwargs) -> typing.Coroutine[typing.Any, typing.Any, R] | R: ...

    @typing.runtime_checkable
    class NotBoundFunction[**P, R](typing.Protocol):
        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> typing.Coroutine[typing.Any, typing.Any, R] | R: ...


class mutation[**P, IntoState: State, BoundState: State | None]:  # noqa: N801
    from_state: State | None

    @typing.overload
    def __init__(
        self: mutation[P, IntoState, BoundState],  # type: ignore
        construct_mutation: BoundMethod[P, BoundState, IntoState],
    ) -> None: ...

    @typing.overload
    def __init__(
        self: mutation[P, IntoState, None],  # type: ignore
        construct_mutation: NotBoundFunction[P, IntoState],
    ) -> None: ...

    def __init__(
        self,
        construct_mutation: typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, IntoState] | IntoState],
    ) -> None:
        self.construct_mutation = construct_mutation
        self.from_state = None

    @typing.overload
    async def __call__(
        self: mutation[P, IntoState, None],
        mutator: StateMutator,
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> IntoState: ...

    @typing.overload
    async def __call__(
        self: mutation[P, IntoState, State],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> IntoState: ...

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> IntoState:
        mutator: StateMutator

        if self.from_state is not None:
            if self.from_state.__mutator__ is None:
                raise RuntimeError("State is not bound. Bind state with .bind(mutator) before mutating")

            await self.from_state.exit()

            new_state = await maybe_awaitable(self.construct_mutation(self.from_state, *args, **kwargs))
            mutator = self.from_state.__mutator__
        else:
            mutator, args = args[0], args[1:]

            if (state := await mutator.get()) is not None:
                await state.exit()

            new_state = await maybe_awaitable(self.construct_mutation(*args, **kwargs))

        new_state.bind(mutator)
        await new_state.enter()

        return new_state

    def __get__(self, instance: State, owner: typing.Any) -> typing.Self:
        if instance is not None:
            self.from_state = instance
        return self


__all__ = ("mutation",)



import dataclasses
import datetime
import pathlib
import sys
import typing
from http import HTTPStatus

import certifi
import rnet
import rnet.exceptions
from rnet import Method as HTTPMethod

from telegrinder.__meta__ import __version__
from telegrinder.client.abc import ABCClient, Response
from telegrinder.modules import json

if typing.TYPE_CHECKING:
    from rnet import ClientConfig, Request

type Data = dict[str, typing.Any] | rnet.Multipart
type Method = typing.Literal["GET", "HEAD", "POST", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]

_METHODS_MAP: typing.Final[dict[Method, HTTPMethod]] = {
    "GET": HTTPMethod.GET,
    "HEAD": HTTPMethod.HEAD,
    "POST": HTTPMethod.POST,
    "PUT": HTTPMethod.PUT,
    "DELETE": HTTPMethod.DELETE,
    "OPTIONS": HTTPMethod.OPTIONS,
    "TRACE": HTTPMethod.TRACE,
    "PATCH": HTTPMethod.PATCH,
}
USER_AGENT: typing.Final = "CPython/{}.{} RNET/3 Telegrinder/{}".format(
    sys.version_info.major,
    sys.version_info.minor,
    __version__,
)
DEFAULT_CONNECTION_TIMEOUT: typing.Final = datetime.timedelta(seconds=30)
DEFAULT_READ_TIMEOUT: typing.Final = datetime.timedelta(seconds=30)
DEFAULT_TIMEOUT: typing.Final = datetime.timedelta(seconds=30)
DEFAULT_HTTP2_MAX_RETRIES: typing.Final = 10
DEFAULT_ZSTD: typing.Final = True
DEFAULT_VERIFY: typing.Final = pathlib.Path(certifi.where())
DEFAULT_ALLOW_REDIRECTS: typing.Final = True
DEFAULT_HTTP2_ONLY: typing.Final = True
DEFAULT_TCP_KEEPALIVE_TIME: typing.Final = datetime.timedelta(seconds=60)
DEFAULT_TCP_KEEPALIVE_INTERVAL: typing.Final = datetime.timedelta(seconds=15)
DEFAULT_TCP_KEEPALIVE_RETRIES: typing.Final = 4
DEFAULT_TCP_USER_TIMEOUT: typing.Final = (
    DEFAULT_TCP_KEEPALIVE_TIME + DEFAULT_TCP_KEEPALIVE_INTERVAL
) * DEFAULT_TCP_KEEPALIVE_RETRIES
DEFAULT_TCP_REUSEADDR: typing.Final = True
DEFAULT_CONNECTION_POOL_IDLE_TIMEOUT: typing.Final = datetime.timedelta(seconds=60)
DEFAULT_CONNECTION_POOL_CONNECTIONS: typing.Final = 32
CONNECTION_POOL_MAX_SIZE: typing.Final = DEFAULT_CONNECTION_POOL_CONNECTIONS * 2


@dataclasses.dataclass(frozen=True, slots=True)
class RnetMultipartBuilder:
    parts: list[rnet.Part] = dataclasses.field(default_factory=list[rnet.Part])

    def add_field(
        self,
        name: str,
        value: typing.Any,
        /,
        *,
        filename: str | None = None,
    ) -> None:
        self.parts.append(rnet.Part(name, value, filename=filename))

    def build(self) -> rnet.Multipart:
        return rnet.Multipart(*self.parts)


class RnetClient(ABCClient):
    __slots__ = ("_timeout", "_client")

    CONNECTION_TIMEOUT_ERRORS: typing.ClassVar = (
        TimeoutError,
        rnet.exceptions.TimeoutError,
        rnet.exceptions.RustPanic,
    )
    CLIENT_CONNECTION_ERRORS: typing.ClassVar = (
        rnet.exceptions.ConnectionError,
        rnet.exceptions.ConnectionResetError,
        rnet.exceptions.TlsError,
        rnet.exceptions.RustPanic,
    )

    def __init__(self, **params: typing.Unpack[ClientConfig]) -> None:
        params.setdefault("user_agent", USER_AGENT)
        params.setdefault("connect_timeout", DEFAULT_CONNECTION_TIMEOUT)
        params.setdefault("read_timeout", DEFAULT_READ_TIMEOUT)
        params.setdefault("verify", DEFAULT_VERIFY)
        params.setdefault("http2_only", DEFAULT_HTTP2_ONLY)
        params.setdefault("zstd", DEFAULT_ZSTD)
        params.setdefault("tcp_keepalive", DEFAULT_TCP_KEEPALIVE_TIME)
        params.setdefault("tcp_keepalive_interval", DEFAULT_TCP_KEEPALIVE_INTERVAL)
        params.setdefault("tcp_keepalive_retries", DEFAULT_TCP_KEEPALIVE_RETRIES)
        params.setdefault("tcp_user_timeout", DEFAULT_TCP_USER_TIMEOUT)
        params.setdefault("tcp_reuse_address", DEFAULT_TCP_REUSEADDR)
        params.setdefault("pool_idle_timeout", DEFAULT_CONNECTION_POOL_IDLE_TIMEOUT)
        params.setdefault("pool_max_idle_per_host", DEFAULT_CONNECTION_POOL_CONNECTIONS)
        params.setdefault("pool_max_size", CONNECTION_POOL_MAX_SIZE)

        self._timeout = params.setdefault("timeout", DEFAULT_TIMEOUT)
        self._client = rnet.Client(**params)

    def __repr__(self) -> str:
        return "<{} {!r}, timeout={!r}>".format(
            type(self).__name__,
            self._client,
            self._timeout,
        )

    @property
    def timeout(self) -> datetime.timedelta:
        return self._timeout

    @classmethod
    def multipart_form_builder(cls) -> RnetMultipartBuilder:
        return RnetMultipartBuilder()

    async def request(
        self,
        url: str,
        method: Method = "GET",
        data: Data | None = None,
        **kwargs: typing.Unpack[Request],
    ) -> Response[rnet.Response]:
        kwargs.setdefault("version", rnet.Version.HTTP_2)
        kwargs.setdefault("zstd", DEFAULT_ZSTD)

        if data is not None:
            if isinstance(data, rnet.Multipart):
                kwargs["multipart"] = data
            elif isinstance(data, dict):
                kwargs["json"] = data

        if (json_body := kwargs.pop("json", None)) is not None:
            kwargs["body"] = json.dumps(json_body)

        if (timeout := kwargs.get("timeout")) is not None and isinstance(timeout, int | float):
            kwargs["timeout"] = datetime.timedelta(seconds=timeout)

        response = await self._client.request(_METHODS_MAP[method], url, **kwargs)
        return Response(
            response=response,
            content=await response.bytes(),
            status=HTTPStatus(response.status.as_int()),
        )

    async def request_text(
        self,
        *,
        url: str,
        method: Method = "GET",
        data: Data | None = None,
        **kwargs: typing.Unpack[Request],
    ) -> str:
        return await (await self.request(url, method, data=data, **kwargs)).response.text_with_charset(encoding="utf-8")

    async def request_bytes(
        self,
        *,
        url: str,
        method: Method = "GET",
        data: Data | None = None,
        **kwargs: typing.Unpack[Request],
    ) -> bytes:
        return (await self.request(url, method, data=data, **kwargs)).content

    async def request_content(
        self,
        *,
        url: str,
        method: Method = "GET",
        data: Data | None = None,
        **kwargs: typing.Unpack[Request],
    ) -> bytes:
        return await self.request_bytes(url=url, method=method, data=data, **kwargs)

    async def request_json(
        self,
        *,
        url: str,
        method: Method = "GET",
        data: Data | None = None,
        **kwargs: typing.Unpack[Request],
    ) -> dict[str, typing.Any]:
        return json.loads(await self.request_bytes(url=url, method=method, data=data, **kwargs))

    async def close(self) -> None:
        return None


__all__ = ("RnetClient", "RnetMultipartBuilder")



import typing

from telegrinder.tools.final import Final


@typing.final
class ParseMode(Final):
    MARKDOWNV2: typing.Final[typing.Literal["MarkdownV2"]] = "MarkdownV2"
    HTML: typing.Final[typing.Literal["HTML"]] = "HTML"


__all__ = ("ParseMode",)



import asyncio
import dataclasses
import datetime
import typing
from functools import partial

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.aio import cancel_future
from telegrinder.tools.global_context.builtin_context import TelegrinderContext

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute
    from telegrinder.bot.dispatch.view.base import View
    from telegrinder.bot.dispatch.waiter_machine.actions import WaiterActions
    from telegrinder.bot.dispatch.waiter_machine.hasher import Hasher
    from telegrinder.bot.dispatch.waiter_machine.machine import HasherWithData

TELEGRINDER_CONTEXT: typing.Final = TelegrinderContext()


def _wrap_async_fn[**P](
    fn: typing.Callable[P, typing.Any],
    /,
) -> typing.Callable[P, None]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        TELEGRINDER_CONTEXT.loop_wrapper.add_task(fn(*args, **kwargs))

    return wrapper


class ShortStateContext[Event: BaseCute[typing.Any] = typing.Any](typing.NamedTuple):
    event: Event
    context: Context


@dataclasses.dataclass
class ShortState[Event: BaseCute[typing.Any] = typing.Any]:
    event: asyncio.Event
    actions: WaiterActions[Event]

    release: ABCRule | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    filter: ABCRule | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    expiration: dataclasses.InitVar[datetime.timedelta] = dataclasses.field(
        kw_only=True,
    )

    expiration_date: datetime.datetime | None = dataclasses.field(init=False, kw_only=True)
    creation_date: datetime.datetime = dataclasses.field(init=False)
    context: ShortStateContext[Event] | None = dataclasses.field(default=None, init=False, kw_only=True)
    drop_timer: asyncio.TimerHandle | None = dataclasses.field(default=None, init=False)

    def __post_init__(self, expiration: datetime.timedelta) -> None:
        self.lifetime = expiration
        self.creation_date = datetime.datetime.now()
        self.expiration_date = (self.creation_date + expiration) if expiration is not None else None

    def schedule_drop[HasherData](
        self,
        dropper: typing.Callable[..., typing.Any],
        /,
        hasher: Hasher[Event, HasherData],
        data: HasherData,
        **context: typing.Any,
    ) -> None:
        self.drop_timer = asyncio.get_running_loop().call_later(
            delay=self.lifetime.total_seconds(),
            callback=partial(_wrap_async_fn(dropper), self, hasher, data, expired=True, **context),
        )

    def schedule_drop_many[ViewType: View, HasherData](
        self,
        dropper: typing.Callable[..., typing.Any],
        /,
        *hashers: HasherWithData[Event, ViewType, HasherData],
        **context: typing.Any,
    ) -> None:
        self.drop_timer = asyncio.get_running_loop().call_later(
            delay=self.lifetime.total_seconds(),
            callback=partial(_wrap_async_fn(dropper), self, *hashers, expired=True, **context),
        )

    def cancel_drop(self) -> None:
        if self.drop_timer is not None:
            self.drop_timer.cancel()
            self.drop_timer = None

    async def cancel(self) -> None:
        waiters = typing.cast(
            "typing.Iterable[asyncio.Future[typing.Any]]",
            self.event._waiters,  # type: ignore
        )

        for future in waiters:
            if not future.cancelled():
                await cancel_future(future)


__all__ = ("ShortState", "ShortStateContext")



import datetime
import typing
from http import HTTPStatus

import msgspec
from kungfu.library.misc import is_ok

from telegrinder.api.api import API
from telegrinder.api.error import APIServerError, InvalidTokenError
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.bot.polling.error_handler import ErrorHandler
from telegrinder.bot.polling.utils import compute_number
from telegrinder.modules import logger
from telegrinder.msgspec_utils import decoder
from telegrinder.types.objects import Update, UpdateType

DEFAULT_OFFSET: typing.Final = 0
DEFAULT_RECONNECT_AFTER: typing.Final = 5.0
DEFAULT_MAX_RECONNECTS: typing.Final = 15


class Polling(ABCPolling):
    __slots__ = (
        "api",
        "timeout",
        "limit",
        "allowed_updates",
        "reconnect_after",
        "max_reconnects",
        "offset",
        "_running",
        "_reconnects_counter",
        "_error_handler",
    )

    def __init__(
        self,
        api: API,
        *,
        timeout: int | float | datetime.timedelta | None = None,
        limit: int | None = None,
        offset: int = DEFAULT_OFFSET,
        reconnect_after: float = DEFAULT_RECONNECT_AFTER,
        max_reconnects: int = DEFAULT_MAX_RECONNECTS,
        include_updates: set[UpdateType] | None = None,
        exclude_updates: set[UpdateType] | None = None,
    ) -> None:
        self.api = api
        self.timeout = timeout if isinstance(timeout, datetime.timedelta) else datetime.timedelta(seconds=timeout or 0)
        self.timeout_seconds = int(self.timeout.total_seconds())
        self.limit = limit
        self.offset = max(DEFAULT_OFFSET, offset)
        self.allowed_updates = self.get_allowed_updates(
            include_updates=include_updates,
            exclude_updates=exclude_updates,
        )
        self.reconnect_after = compute_number(DEFAULT_RECONNECT_AFTER, reconnect_after, 0.0)
        self.max_reconnects = compute_number(DEFAULT_MAX_RECONNECTS, max_reconnects, 0)
        self._running = False
        self._reconnects_counter = 0
        self._error_handler = ErrorHandler(self)

    def __repr__(self) -> str:
        return (
            "<{}: api={!r}, running={}, offset={}, timeout={}, limit={}, "
            "allowed_updates={!r}, max_reconnects={}, reconnect_after={}>"
        ).format(
            type(self).__name__,
            self.api,
            self._running,
            self.offset,
            self.timeout,
            self.limit,
            self.allowed_updates,
            self.max_reconnects,
            self.reconnect_after,
        )

    @staticmethod
    def get_allowed_updates(
        *,
        include_updates: set[UpdateType] | None = None,
        exclude_updates: set[UpdateType] | None = None,
    ) -> list[UpdateType]:
        allowed_updates = list(UpdateType)

        if include_updates and exclude_updates:
            return [x for x in allowed_updates if x in include_updates and x not in exclude_updates]

        if exclude_updates:
            return [x for x in allowed_updates if x not in exclude_updates]

        if include_updates:
            return [x for x in allowed_updates if x in include_updates]

        return allowed_updates

    @property
    def reconnects_counter(self) -> int:
        return self._reconnects_counter

    @property
    def running(self) -> bool:
        return self._running

    def _reset_reconnects_counter(self) -> None:
        self._reconnects_counter = 0

    async def get_updates(self) -> msgspec.Raw:
        try:
            raw_updates = await self.api.request_raw(
                method="getUpdates",
                data=dict(
                    offset=self.offset,
                    limit=self.limit,
                    timeout=self.timeout_seconds,
                    allowed_updates=self.allowed_updates,
                ),
                timeout=self.timeout + self.api.http.timeout,
            )
        except TimeoutError:
            return msgspec.Raw(b"")

        if is_ok(raw_updates):
            return raw_updates.value

        match (error := raw_updates.error).status_code:
            case HTTPStatus.CONFLICT:
                error = APIServerError("Cannot run polling, because another bot instance is already running.")
            case HTTPStatus.UNAUTHORIZED | HTTPStatus.NOT_FOUND as status:
                error = InvalidTokenError(
                    "Token seems to be invalid" if status == HTTPStatus.NOT_FOUND else "Invalid token (unauthorized)",
                )
            case HTTPStatus.BAD_GATEWAY | HTTPStatus.GATEWAY_TIMEOUT as status:
                error = APIServerError(
                    message="Telegram API server responded a {}".format(status.name.replace("_", " ")),
                    retry_after=int(self.reconnect_after),
                )

        raise error from None

    async def listen(self) -> typing.AsyncGenerator[list[Update], None]:
        await logger.adebug("Listening polling")
        self._running = True

        with decoder(list[Update]) as updates_decoder:
            while self._running:
                try:
                    if (raw := await self.get_updates()) and (updates := updates_decoder.decode(raw)):
                        yield updates
                        self.offset = updates[-1].update_id + 1

                    if self._reconnects_counter != 0:
                        self._reset_reconnects_counter()
                except BaseException as error:
                    if not await self._error_handler.handle(error):
                        await logger.aexception("Traceback message below:")

                    if isinstance(error, self.api.http.CONNECTION_TIMEOUT_ERRORS):
                        self._reconnects_counter += 1

    def stop(self) -> None:
        self._running = False
        self._reset_reconnects_counter()


__all__ = ("Polling",)



import abc
import typing

from telegrinder.types.objects import InlineKeyboardMarkup, ReplyKeyboardMarkup

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.button import BaseButton

from telegrinder.tools.keyboard.utils import RowButtons, copy_keyboard

type DictStrAny = dict[str, typing.Any]
type AnyMarkup = ReplyKeyboardMarkup | InlineKeyboardMarkup
type RawKeyboard = list[list[DictStrAny]]
type Button = DictStrAny | list[DictStrAny] | BaseButton | RowButtons[BaseButton]


class ABCKeyboard(typing.Protocol):
    keyboard: RawKeyboard

    __button_class__: typing.ClassVar[type[BaseButton]]

    @abc.abstractmethod
    def dict(self) -> DictStrAny:
        pass

    @abc.abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass

    @abc.abstractmethod
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        pass

    def add(self, button: Button, /) -> typing.Self:
        if not self.keyboard:
            self.row()

        if isinstance(button, RowButtons):
            self.keyboard[-1].extend(button.get_data())
            if button.auto_row:
                self.row()
            return self

        if isinstance(button, list):
            self.keyboard[-1].extend(button)
            return self

        self.keyboard[-1].append(button if isinstance(button, dict) else button.get_data())
        return self

    def row(self) -> typing.Self:
        if len(self.keyboard) and not len(self.keyboard[-1]):
            return self

        self.keyboard.append([])
        return self

    def format_text(self, **format_data: typing.Any) -> typing.Self:
        copy_keyboard = self.copy()

        for row in self.keyboard:
            for button in row:
                button.update(dict(text=button["text"].format(**format_data)))

        return copy_keyboard

    def merge(self, other: typing.Self, /) -> typing.Self:
        self.keyboard.extend(copy_keyboard(other.keyboard))
        return self

    def merge_to_last_row(self, other: typing.Self, /) -> typing.Self:
        total_rows = len(other.keyboard)

        for index, row in enumerate(copy_keyboard(other.keyboard), start=1):
            for button in row:
                self.keyboard[-1].append(button)

            if index < total_rows:
                self.keyboard.append([])

        return self


__all__ = ("ABCKeyboard",)



import pytest

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.types.objects import Update

from .node_scope import make_node_scope


def make_context(api: API, update: Update) -> Context:
    scope = make_node_scope(api, update)
    ctx = Context()
    ctx.add_roots(api, update, scope)
    return ctx


@pytest.fixture()
def context_factory(api_instance: API):
    def factory(update: Update) -> Context:
        return make_context(api_instance, update)

    return factory


@pytest.fixture()
def message_context(api_instance: API, message_update: Update) -> Context:
    return make_context(api_instance, message_update)


@pytest.fixture()
def callback_query_context(api_instance: API, callback_query_update: Update) -> Context:
    return make_context(api_instance, callback_query_update)



import dataclasses
import typing

import msgspec
from nodnod.error import NodeError
from nodnod.interface.generic import generic_node

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.msgspec_utils import decoder
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Model, Update

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

type DataclassType = DataclassInstance | msgspec.Struct | dict[str, typing.Any]


@generic_node
class EventNode[Dataclass: DataclassType]:  # type: ignore[reportRedeclaration]
    @classmethod
    def __compose__(
        cls,
        api: API,
        raw_update: Update,
        dataclass: type[Dataclass],
        update_cute: UpdateCute,
    ) -> DataclassType:
        orig_dataclass = typing.get_origin(dataclass) or dataclass

        if orig_dataclass is UpdateCute:
            return update_cute

        if issubclass(orig_dataclass, BaseCute | Model):
            incoming_update = (
                update_cute.incoming_update if issubclass(orig_dataclass, BaseCute) else raw_update.incoming_update
            )

            if type(incoming_update) is not orig_dataclass:
                raise NodeError(f"Incoming update is not `{fullname(orig_dataclass)}`.")

            return incoming_update

        try:
            if issubclass(orig_dataclass, msgspec.Struct) or dataclasses.is_dataclass(
                orig_dataclass,
            ):
                obj = decoder.convert(
                    obj=raw_update.incoming_update,
                    type=dataclass,
                    from_attributes=True,
                )
            else:
                obj = dataclass(**raw_update.incoming_update.to_full_dict())

            return obj
        except Exception as exc:
            raise NodeError(
                f"Cannot parse an update object into `{fullname(dataclass)}`",
                from_error=NodeError(exc),
            )


if typing.TYPE_CHECKING:
    type EventNode[Dataclass: DataclassType] = Dataclass


__all__ = ("EventNode",)



import typing
from abc import ABC
from functools import cached_property

from kungfu.library.monad.result import Error, Ok
from nodnod.agent.event_loop.agent import EventLoopAgent
from nodnod.error import NodeError
from nodnod.interface.node_from_function import create_node_from_function

from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.compose import compose, create_composable
from telegrinder.node.utils import get_globals_from_function, get_locals_from_function
from telegrinder.tools.fullname import fullname
from telegrinder.tools.lifespan import Lifespan

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

    from telegrinder.node.compose import Composable

type Node = typing.Any
type MiddlewareResult = bool | typing.Awaitable[bool | None] | None


async def run_pre_middleware(middleware: ABCMiddleware, context: Context) -> bool:
    if middleware.is_pre:
        return bool(await run_middleware(middleware, context, composable=middleware.pre_composable))
    return True


async def run_post_middleware(middleware: ABCMiddleware, context: Context) -> None:
    if middleware.is_post:
        await run_middleware(middleware, context, composable=middleware.post_composable)


async def run_middleware(
    middleware: ABCMiddleware,
    context: Context,
    *,
    composable: Composable[typing.Any],
) -> bool | None:
    async with compose(composable, context) as result:
        match result:
            case Ok(response):
                return response
            case Error(error):
                await logger.adebug(
                    "Middleware `{!r}` failed with error:{}\n",
                    middleware_name := fullname(middleware),
                    NodeError(f"* failed to compose middleware `{middleware_name}`", from_error=error),
                )


class ABCMiddleware(ABC):
    agent_cls: type[Agent] = EventLoopAgent
    pre_required_nodes: typing.Mapping[str, Node] | None = None
    post_required_nodes: typing.Mapping[str, Node] | None = None

    def __repr__(self) -> str:
        return "<{}{}>".format(
            "".join(
                (
                    f"{'pre-' if self.is_pre else ''}",
                    f"{'post-' if self.is_post else ''}",
                ),
            ),
            f"middleware `{fullname(self)}`",
        )

    @property
    def is_pre(self) -> bool:
        return type(self).pre is not ABCMiddleware.pre

    @property
    def is_post(self) -> bool:
        return type(self).post is not ABCMiddleware.post

    @cached_property
    def pre_composable(self) -> Composable:
        return self.get_composable(self.pre, self.agent_cls, required_nodes=self.pre_required_nodes)

    @cached_property
    def post_composable(self) -> Composable:
        return self.get_composable(self.post, self.agent_cls, required_nodes=self.post_required_nodes)

    @staticmethod
    def get_composable(
        method: typing.Callable[..., typing.Any],
        agent_cls: type[Agent] | None,
        required_nodes: typing.Mapping[str, Node] | None,
    ) -> Composable:
        node = create_node_from_function(
            method,
            dependencies=required_nodes,
            forward_refs=get_globals_from_function(method),
            namespace=get_locals_from_function(method),
        )
        return create_composable(node, agent_cls=agent_cls)

    def pre(self, *args: typing.Any, **kwargs: typing.Any) -> MiddlewareResult: ...

    def post(self, *args: typing.Any, **kwargs: typing.Any) -> MiddlewareResult: ...

    def to_lifespan(self, context: Context) -> Lifespan:
        return Lifespan(
            startup_tasks=[run_pre_middleware(self, context)],
            shutdown_tasks=[run_post_middleware(self, context)],
        )


__all__ = ("ABCMiddleware", "run_post_middleware", "run_pre_middleware")



import dataclasses
import typing
from functools import cached_property
from types import MappingProxyType

from telegrinder.bot.dispatch.return_manager.callback_query import CallbackQueryReturnManager
from telegrinder.bot.dispatch.return_manager.inline_query import InlineQueryReturnManager
from telegrinder.bot.dispatch.return_manager.message import MessageReturnManager
from telegrinder.bot.dispatch.return_manager.pre_checkout_query import PreCheckoutQueryReturnManager
from telegrinder.bot.dispatch.view.base import (
    ErrorView,
    EventView,
    RawEventView,
    View,
)
from telegrinder.bot.dispatch.view.media_group import MediaGroupView
from telegrinder.types.enums import UpdateType

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager

EXCLUDE_VIEW_META: typing.Final = dict(exclude_view=True)


def event_view[T: EventView](
    update_type: UpdateType,
    return_manager: ABCReturnManager | None = None,
    /,
) -> typing.Callable[[], T]:
    def factory() -> T:
        return typing.cast("T", EventView(update_type, return_manager))

    return factory


def view[T: View](
    view_class: type[View],
    return_manager: ABCReturnManager | None = None,
    /,
) -> typing.Callable[[], T]:
    def factory() -> T:
        return typing.cast("T", view_class(return_manager=return_manager))

    return factory


event_model_view = view


@dataclasses.dataclass(kw_only=True)
class EventViewBox[
    MessageView: EventView = EventView,
    EditedMessageView: EventView = EventView,
    ChannelPostView: EventView = EventView,
    EditedChannelPostView: EventView = EventView,
    BusinessConnectionView: EventView = EventView,
    BusinessMessageView: EventView = EventView,
    EditedBusinessMessageView: EventView = EventView,
    DeletedBusinessMessagesView: EventView = EventView,
    MessageReactionView: EventView = EventView,
    MessageReactionCountView: EventView = EventView,
    InlineQueryView: EventView = EventView,
    ChosenInlineResultView: EventView = EventView,
    CallbackQueryView: EventView = EventView,
    ShippingQueryView: EventView = EventView,
    PreCheckoutQueryView: EventView = EventView,
    PurchasedPaidMediaView: EventView = EventView,
    PollView: EventView = EventView,
    PollAnswerView: EventView = EventView,
    MyChatMemberView: EventView = EventView,
    ChatMemberView: EventView = EventView,
    ChatJoinRequestView: EventView = EventView,
    ChatBoostView: EventView = EventView,
    RemovedChatBoostView: EventView = EventView,
]:
    message: MessageView = dataclasses.field(default_factory=event_view(UpdateType.MESSAGE, MessageReturnManager()))
    edited_message: EditedMessageView = dataclasses.field(
        default_factory=event_view(UpdateType.EDITED_MESSAGE, MessageReturnManager()),
    )
    channel_post: ChannelPostView = dataclasses.field(
        default_factory=event_view(UpdateType.CHANNEL_POST, MessageReturnManager()),
    )
    edited_channel_post: EditedChannelPostView = dataclasses.field(
        default_factory=event_view(UpdateType.EDITED_CHANNEL_POST, MessageReturnManager()),
    )
    business_connection: BusinessConnectionView = dataclasses.field(
        default_factory=event_view(UpdateType.BUSINESS_CONNECTION),
    )
    business_message: BusinessMessageView = dataclasses.field(
        default_factory=event_view(UpdateType.BUSINESS_MESSAGE, MessageReturnManager()),
    )
    edited_business_message: EditedBusinessMessageView = dataclasses.field(
        default_factory=event_view(UpdateType.EDITED_BUSINESS_MESSAGE, MessageReturnManager()),
    )
    deleted_business_messages: DeletedBusinessMessagesView = dataclasses.field(
        default_factory=event_view(UpdateType.DELETED_BUSINESS_MESSAGES),
    )
    message_reaction: MessageReactionView = dataclasses.field(default_factory=event_view(UpdateType.MESSAGE_REACTION))
    message_reaction_count: MessageReactionCountView = dataclasses.field(
        default_factory=event_view(UpdateType.MESSAGE_REACTION_COUNT),
    )
    inline_query: InlineQueryView = dataclasses.field(
        default_factory=event_view(UpdateType.INLINE_QUERY, InlineQueryReturnManager()),
    )
    chosen_inline_result: ChosenInlineResultView = dataclasses.field(
        default_factory=event_view(UpdateType.CHOSEN_INLINE_RESULT),
    )
    callback_query: CallbackQueryView = dataclasses.field(
        default_factory=event_view(UpdateType.CALLBACK_QUERY, CallbackQueryReturnManager()),
    )
    shipping_query: ShippingQueryView = dataclasses.field(default_factory=event_view(UpdateType.SHIPPING_QUERY))
    pre_checkout_query: PreCheckoutQueryView = dataclasses.field(
        default_factory=event_view(UpdateType.PRE_CHECKOUT_QUERY, PreCheckoutQueryReturnManager()),
    )
    purchased_paid_media: PurchasedPaidMediaView = dataclasses.field(
        default_factory=event_view(UpdateType.PURCHASED_PAID_MEDIA),
    )
    poll: PollView = dataclasses.field(default_factory=event_view(UpdateType.POLL))
    poll_answer: PollAnswerView = dataclasses.field(default_factory=event_view(UpdateType.POLL_ANSWER))
    my_chat_member: MyChatMemberView = dataclasses.field(default_factory=event_view(UpdateType.MY_CHAT_MEMBER))
    chat_member: ChatMemberView = dataclasses.field(default_factory=event_view(UpdateType.CHAT_MEMBER))
    chat_join_request: ChatJoinRequestView = dataclasses.field(default_factory=event_view(UpdateType.CHAT_JOIN_REQUEST))
    chat_boost: ChatBoostView = dataclasses.field(default_factory=event_view(UpdateType.CHAT_BOOST))
    removed_chat_boost: RemovedChatBoostView = dataclasses.field(
        default_factory=event_view(UpdateType.REMOVED_CHAT_BOOST),
    )


@dataclasses.dataclass(kw_only=True)
class EventModelViewBox[MediaGroup: View = MediaGroupView]:
    media_group: MediaGroup = dataclasses.field(
        default_factory=event_model_view(MediaGroupView, MessageReturnManager())
    )


@dataclasses.dataclass(kw_only=True)
class ViewBox[
    MessageView: EventView = EventView,
    EditedMessageView: EventView = EventView,
    ChannelPostView: EventView = EventView,
    EditedChannelPostView: EventView = EventView,
    BusinessConnectionView: EventView = EventView,
    BusinessMessageView: EventView = EventView,
    EditedBusinessMessageView: EventView = EventView,
    DeletedBusinessMessagesView: EventView = EventView,
    MessageReactionView: EventView = EventView,
    MessageReactionCountView: EventView = EventView,
    InlineQueryView: EventView = EventView,
    ChosenInlineResultView: EventView = EventView,
    CallbackQueryView: EventView = EventView,
    ShippingQueryView: EventView = EventView,
    PreCheckoutQueryView: EventView = EventView,
    PurchasedPaidMediaView: EventView = EventView,
    PollView: EventView = EventView,
    PollAnswerView: EventView = EventView,
    MyChatMemberView: EventView = EventView,
    ChatMemberView: EventView = EventView,
    ChatJoinRequestView: EventView = EventView,
    ChatBoostView: EventView = EventView,
    RemovedChatBoostView: EventView = EventView,
    MediaGroup: View = MediaGroupView,
    Error: ErrorView = ErrorView,
    RawEvent: RawEventView = RawEventView,
](
    EventModelViewBox[MediaGroup],
    EventViewBox[
        MessageView,
        EditedMessageView,
        ChannelPostView,
        EditedChannelPostView,
        BusinessConnectionView,
        BusinessMessageView,
        EditedBusinessMessageView,
        DeletedBusinessMessagesView,
        MessageReactionView,
        MessageReactionCountView,
        InlineQueryView,
        ChosenInlineResultView,
        CallbackQueryView,
        ShippingQueryView,
        PreCheckoutQueryView,
        PurchasedPaidMediaView,
        PollView,
        PollAnswerView,
        MyChatMemberView,
        ChatMemberView,
        ChatJoinRequestView,
        ChatBoostView,
        RemovedChatBoostView,
    ],
):
    event_error: Error = dataclasses.field(default_factory=view(ErrorView), metadata=EXCLUDE_VIEW_META)
    raw: RawEvent = dataclasses.field(default_factory=view(RawEventView), metadata=EXCLUDE_VIEW_META)

    @cached_property
    def views(self) -> MappingProxyType[str, View]:
        return MappingProxyType(
            mapping={
                field.name: obj
                for field in dataclasses.fields(self)
                if isinstance(obj := getattr(self, field.name), View)
                and field.metadata.get("exclude_view", False) is False
            },
        )


__all__ = ("EventModelViewBox", "EventViewBox", "ViewBox")



import typing

from telegrinder.api.api import API
from telegrinder.bot.dispatch import dispatch as dp
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.polling import polling as pg
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.modules import logger
from telegrinder.tools.global_context.builtin_context import TelegrinderContext
from telegrinder.tools.loop_wrapper import LoopWrapper

TELEGRINDER_CONTEXT: typing.Final = TelegrinderContext()


class Telegrinder[Dispatch: ABCDispatch = dp.Dispatch, Polling: ABCPolling = pg.Polling]:
    def __init__(
        self,
        api: API,
        *,
        dispatch: Dispatch | None = None,
        polling: Polling | None = None,
        loop_wrapper: LoopWrapper | None = None,
    ) -> None:
        self.api = api
        self.dispatch = typing.cast("Dispatch", dispatch or dp.Dispatch())
        self.polling = typing.cast("Polling", polling or pg.Polling(api))
        self.loop_wrapper = loop_wrapper or TELEGRINDER_CONTEXT.loop_wrapper

    def __repr__(self) -> str:
        return "<{}: api={!r}, dispatch={!r}, polling={!r}, loop_wrapper={!r}>".format(
            type(self).__name__,
            self.api,
            self.dispatch,
            self.polling,
            self.loop_wrapper,
        )

    @property
    def on(self) -> Dispatch:
        return self.dispatch

    async def drop_pending_updates(self) -> None:
        await logger.adebug("Dropping pending updates")
        await self.api.delete_webhook(drop_pending_updates=True)

    async def run_polling(
        self,
        *,
        offset: int = 0,
        skip_updates: bool = False,
    ) -> None:
        self.polling.offset = offset

        async def listen_polling() -> None:
            if skip_updates:
                await self.drop_pending_updates()

            async for updates in self.polling.listen():
                for update in updates:
                    self.loop_wrapper.add_task(self.dispatch.feed(self.api, update))

        self.loop_wrapper.add_task(listen_polling())

    def run_forever(self, *, offset: int = 0, skip_updates: bool = False) -> typing.NoReturn:
        logger.info("Running blocking polling (id={})", self.api.id)
        self.loop_wrapper.add_task(self.run_polling(offset=offset, skip_updates=skip_updates))
        self.loop_wrapper.run()


__all__ = ("Telegrinder",)



"""The example uses the aiosqlite driver. It is recommended to use asynchronous drivers
to work with databases.
"""

import aiosqlite  # type: ignore
from kungfu.library.monad.option import Nothing

from telegrinder import API, ABCMiddleware, Message, Telegrinder, Token
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.msgspec_utils import Option, decoder
from telegrinder.rules import IsReply, Markup, MessageEntities, Text
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import User

db_path = "examples/assets/users.db"
bot = Telegrinder(API(Token.from_env()))


@bot.loop_wrapper.lifespan.on_startup
async def create_table() -> None:
    async with aiosqlite.connect(db_path) as conn:  # noqa: SIM117
        async with conn.cursor() as cur:
            await cur.execute(
                "create table if not exists users("
                "id integer unique,"
                "is_bot boolean,"
                "is_premium boolean,"
                "first_name varchar(64),"
                "last_name varchar(64),"
                "username varchar(32),"
                "language_code varchar(5))"
            )
            await conn.commit()


def get_result_with_names(
    cursor: aiosqlite.Cursor,
    row: aiosqlite.Row,
    as_bool: set[str] | None = None,
) -> dict:
    column_names = [d[0] for d in cursor.description]
    resulting_row = {}
    for index, column_name in enumerate(column_names):
        value = row[index]
        resulting_row[column_name] = bool(value) if column_name in (as_bool or ()) and value in (0, 1) else value
    return resulting_row


class DummyDatabase:
    async def set_user(self, user: User) -> None:
        async with aiosqlite.connect(db_path) as conn:
            async with conn.cursor() as cur:
                await cur.execute(
                    """
                    INSERT OR REPLACE INTO users(
                        id, is_bot, first_name, last_name,
                        username, language_code, is_premium
                    ) VALUES (
                        :id, :is_bot, :first_name, :last_name,
                        :username, :language_code, :is_premium
                    );
                    """,
                    user.to_full_dict(),
                )
            await conn.commit()

    async def get_user_by_username(self, username: str) -> Option[User]:
        async with (
            aiosqlite.connect(db_path) as conn,
            conn.cursor() as cur,
        ):
            row = await (await cur.execute("SELECT * FROM users WHERE username = ?", (username,))).fetchone()
            if row is None:
                return Nothing()
            return decoder.convert(
                obj=get_result_with_names(cur, row, as_bool={"is_premium", "is_bot"}),
                type=Option[User],  # type: ignore
            )

    async def get_or_set_user(self, user: User) -> User:
        if (await self.get_user_by_id(user.id)).unwrap_or_none() is None:
            await self.set_user(user)
        return user

    async def get_user_by_id(self, user_id: int) -> Option[User]:
        async with (
            aiosqlite.connect(db_path) as conn,
            conn.cursor() as cur,
        ):
            row = await (await cur.execute("SELECT * FROM users WHERE id = ?", (user_id,))).fetchone()
            if row is None:
                return Nothing()
            return decoder.convert(
                obj=get_result_with_names(cur, row, as_bool={"is_premium", "is_bot"}),
                type=Option[User],  # type: ignore
            )


db = DummyDatabase()


@bot.on.message.register_middleware
class UserRegistrarMiddleware(ABCMiddleware):
    async def pre(self, event: Message) -> bool:
        if event.from_:
            await db.set_user(event.from_user)
        if event.reply_to_message and event.reply_to_message.unwrap().from_:
            await db.set_user(event.reply_to_message.unwrap().from_user)
        return True


class MentionRule(
    ABCRule,
    requires=[
        Text("/get_user") & IsReply()
        | Markup("/get_user id<user_id:int>")
        | MessageEntities(MessageEntityType.TEXT_MENTION) & Markup("/get_user <!>")
        | MessageEntities(MessageEntityType.MENTION) & Markup("/get_user @<username>")
    ],
):
    async def check(self, message: Message, ctx: Context) -> bool:
        user = None

        match ctx:
            case {"message_entities": message_entities} if message_entities[0].type == MessageEntityType.TEXT_MENTION:
                user = await db.get_or_set_user(message_entities[0].user.unwrap())
            case {"username": username}:
                user = (await db.get_user_by_username(username)).unwrap_or_none()
            case {"user_id": user_id}:
                user = (await db.get_user_by_id(user_id)).unwrap_or_none()
            case _ if message.reply_to_message:
                user = message.reply_to_message.unwrap().from_.unwrap_or_none()

        if user is None:
            return False

        ctx.set("mentioned_user", user)
        return True


@bot.on.message(MentionRule())
async def get_user(mentioned_user: User) -> str:
    return f"""
    Id -> {mentioned_user.id}
    Is bot -> {"✅" if mentioned_user.is_bot else "❌"}
    First name -> {mentioned_user.first_name}
    Last name -> {mentioned_user.last_name.unwrap_or(repr("UNKNOWN"))}
    Is premium -> {"✅" if mentioned_user.is_premium.unwrap_or(False) else "❌"}
    Language code -> {mentioned_user.language_code.unwrap_or(repr("UNKNOWN"))}
    """.replace("    ", "")


bot.run_forever()



from telegrinder.tools.keyboard.abc import ABCKeyboard
from telegrinder.tools.keyboard.button import Button, InlineButton
from telegrinder.tools.keyboard.keyboard import InlineKeyboard, Keyboard
from telegrinder.tools.keyboard.utils import RowButtons

__all__ = ("ABCKeyboard", "Button", "InlineButton", "InlineKeyboard", "Keyboard", "RowButtons")



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import BusinessConnection


class BusinessConnectionCute(BaseCute[BusinessConnection], BusinessConnection, kw_only=True):
    pass


__all__ = ("BusinessConnectionCute",)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.business_connection import BusinessConnectionCute
from telegrinder.bot.cute_types.business_messages_deleted import BusinessMessagesDeletedCute
from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.chat_boost_removed import ChatBoostRemovedCute
from telegrinder.bot.cute_types.chat_boost_updated import ChatBoostUpdatedCute
from telegrinder.bot.cute_types.chat_join_request import ChatJoinRequestCute
from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.cute_types.chosen_inline_result import ChosenInlineResultCute
from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.message_reaction_count_updated import MessageReactionCountUpdatedCute
from telegrinder.bot.cute_types.message_reaction_updated import MessageReactionUpdatedCute
from telegrinder.bot.cute_types.paid_media_purchased import PaidMediaPurchasedCute
from telegrinder.bot.cute_types.poll import PollCute
from telegrinder.bot.cute_types.poll_answer import PollAnswerCute
from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.bot.cute_types.shipping_query import ShippingQueryCute
from telegrinder.bot.cute_types.update import UpdateCute

__all__ = (
    "BaseCute",
    "BusinessConnectionCute",
    "BusinessMessagesDeletedCute",
    "CallbackQueryCute",
    "ChatBoostRemovedCute",
    "ChatBoostUpdatedCute",
    "ChatJoinRequestCute",
    "ChatMemberUpdatedCute",
    "ChosenInlineResultCute",
    "InlineQueryCute",
    "MessageCute",
    "MessageReactionCountUpdatedCute",
    "MessageReactionUpdatedCute",
    "PaidMediaPurchasedCute",
    "PollAnswerCute",
    "PollCute",
    "PreCheckoutQueryCute",
    "ShippingQueryCute",
    "UpdateCute",
)



from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware, run_post_middleware, run_pre_middleware
from telegrinder.bot.dispatch.middleware.box import MiddlewareBox
from telegrinder.bot.dispatch.middleware.filter import FilterMiddleware
from telegrinder.bot.dispatch.middleware.media_group import MediaGroupMiddleware

__all__ = (
    "ABCMiddleware",
    "FilterMiddleware",
    "MediaGroupMiddleware",
    "MiddlewareBox",
    "run_post_middleware",
    "run_pre_middleware",
)



import dataclasses
import enum
import typing

from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.types.objects import (
    ChatMemberAdministrator,
    ChatMemberBanned,
    ChatMemberLeft,
    ChatMemberMember,
    ChatMemberOwner,
    ChatMemberRestricted,
)

type _MemberStatus = MemberStatusFlag | MemberStatusUpdating
type ChatMember = typing.Union[
    ChatMemberAdministrator,
    ChatMemberBanned,
    ChatMemberLeft,
    ChatMemberMember,
    ChatMemberOwner,
    ChatMemberRestricted,
]
type ChatMemberUpdated = ChatMemberUpdatedCute


@dataclasses.dataclass(frozen=True, slots=True)
class MemberStatusUpdating:
    old: MemberStatusFlag
    new: MemberStatusFlag

    def __invert__(self) -> typing.Self:
        return dataclasses.replace(self, old=self.new, new=self.old)


class MemberStatusFlag(enum.IntFlag):
    CREATOR = enum.auto()
    ADMINISTRATOR = enum.auto()
    MEMBER = enum.auto()
    LEFT = enum.auto()
    KICKED = enum.auto()
    RESTRICTED = enum.auto()
    RESTRICTED_NOT_MEMBER = enum.auto()

    RESTRICTED_MEMBER = MEMBER | RESTRICTED

    IS_MEMBER = CREATOR | ADMINISTRATOR | MEMBER | RESTRICTED
    IS_NOT_MEMBER = LEFT | KICKED | RESTRICTED_NOT_MEMBER

    IS_ADMIN = CREATOR | ADMINISTRATOR
    IS_NOT_ADMIN = RESTRICTED_MEMBER | IS_NOT_MEMBER

    def __rshift__(self, other: object, /) -> MemberStatusUpdating:
        if not isinstance(other, type(self)):
            return NotImplemented
        return MemberStatusUpdating(old=self, new=other)

    def __lshift__(self, other: object, /) -> MemberStatusUpdating:
        if not isinstance(other, type(self)):
            return NotImplemented
        return MemberStatusUpdating(old=other, new=self)


class MemberStatus:
    CREATOR = MemberStatusFlag.CREATOR
    ADMINISTRATOR = MemberStatusFlag.ADMINISTRATOR
    MEMBER = MemberStatusFlag.MEMBER
    LEFT = MemberStatusFlag.LEFT
    KICKED = MemberStatusFlag.KICKED
    RESTRICTED = MemberStatusFlag.RESTRICTED_MEMBER | MemberStatusFlag.RESTRICTED_NOT_MEMBER

    RESTRICTED_MEMBER = MemberStatusFlag.RESTRICTED_MEMBER
    RESTRICTED_NOT_MEMBER = MemberStatusFlag.RESTRICTED_NOT_MEMBER

    IS_MEMBER = MemberStatusFlag.IS_MEMBER
    IS_NOT_MEMBER = ~IS_MEMBER

    IS_ADMIN = MemberStatusFlag.IS_ADMIN
    IS_NOT_ADMIN = ~IS_ADMIN

    JOIN = IS_NOT_MEMBER >> IS_MEMBER
    LEAVE = ~JOIN
    PROMOTED = (RESTRICTED_MEMBER | IS_NOT_MEMBER) >> ADMINISTRATOR


class ChatMemberUpdatedRule(ABCRule):
    def __init__(self, status: _MemberStatus, /) -> None:
        self.status = status

    def check_status(
        self,
        member: ChatMember,
        status: MemberStatusFlag,
        checking_status: MemberStatusFlag,
    ) -> bool:
        if (
            MemberStatusFlag.RESTRICTED_MEMBER in checking_status
            or MemberStatusFlag.RESTRICTED_NOT_MEMBER in checking_status
        ) and status == MemberStatusFlag.RESTRICTED:
            return (
                True
                if MemberStatus.RESTRICTED in checking_status
                else (getattr(member, "is_member", None) is (MemberStatusFlag.RESTRICTED_MEMBER in checking_status))
            )

        return status == checking_status or status in checking_status

    def check(self, event: ChatMemberUpdated) -> bool:
        checking_new_status, checking_old_status = (
            (
                self.status.new,
                self.status.old,
            )
            if isinstance(self.status, MemberStatusUpdating)
            else (self.status, None)
        )
        new_member, old_member = event.new_chat_member.v, event.old_chat_member.v
        new_member_status, old_member_status = (
            MEMBER_STATUS_MAP[new_member.status],
            MEMBER_STATUS_MAP[old_member.status],
        )
        return (
            all(
                (
                    self.check_status(old_member, old_member_status, checking_old_status),
                    self.check_status(new_member, new_member_status, checking_new_status),
                ),
            )
            if checking_old_status is not None
            else self.check_status(new_member, new_member_status, checking_new_status)
        )


MEMBER_STATUS_MAP: typing.Final = {
    "creator": MemberStatusFlag.CREATOR,
    "administrator": MemberStatusFlag.ADMINISTRATOR,
    "member": MemberStatusFlag.MEMBER,
    "left": MemberStatusFlag.LEFT,
    "kicked": MemberStatusFlag.KICKED,
    "restricted": MemberStatusFlag.RESTRICTED,
}


__all__ = ("ChatMemberUpdatedRule", "MemberStatus")



from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node

from telegrinder.api.api import API
from telegrinder.node.scope import global_node
from telegrinder.types.objects import User


@global_node
@scalar_node
class Me:
    @classmethod
    async def __compose__(cls, api: API) -> User:
        me = await api.get_me()
        return me.expect(NodeError("Can't complete api.get_me() request."))


@global_node
@scalar_node
class BotUsername:
    @classmethod
    def __compose__(cls, me: Me) -> str:
        return me.username.unwrap()


__all__ = ("BotUsername", "Me")



from telegrinder import API, Bot, Message, Token, setup_logger
from telegrinder.node import UserSource
from telegrinder.rules import Magic

setup_logger(level="DEBUG")
bot = Bot(API(Token.from_env()))


@bot.on.message(
    Magic(Message.photo).expect().then(lambda sizes: sizes[-1]).ensure(lambda size: size.width > size.height)
)
async def cine_handler():
    return "cine"


@bot.on.message(Magic(Message.chat.title).expect().then(lambda title: "friend" in title.lower()))
async def friend_chat_handler():
    return "friend chat"


@bot.on.message(Magic(UserSource).ensure(lambda src: src.first_name == "arseny"))
async def user_handler():
    return "hello arseny"


bot.run_forever(skip_updates=True)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import ChatBoostRemoved


class ChatBoostRemovedCute(BaseCute[ChatBoostRemoved], ChatBoostRemoved, kw_only=True):
    pass


__all__ = ("ChatBoostRemovedCute",)



import pathlib
import typing

import msgspec

from .types.methods import GeneratorMethods
from .types.objects import GeneratorObjects


def dec_hook(x: type[typing.Any], obj: typing.Any) -> typing.Any:
    if issubclass(x, pathlib.Path) and isinstance(obj, str | pathlib.Path):
        return x(obj)
    raise TypeError


class TelegramBotAPI(msgspec.Struct):
    version: str
    schema_url: str

    @property
    def version_number(self) -> float:
        return float(self.version.removeprefix("v"))


class TypedDefaultParameter(msgspec.Struct):
    name: str
    type: str


class GeneratorAPI(msgspec.Struct):
    typed_default_parameters: list[TypedDefaultParameter] = msgspec.field(
        name="typed-default-parameters",
        default_factory=list,
    )


class Generator(msgspec.Struct):
    directory: pathlib.Path
    api: GeneratorAPI = msgspec.field(default_factory=lambda: GeneratorAPI())
    objects: GeneratorObjects = msgspec.field(default_factory=lambda: GeneratorObjects())
    methods: GeneratorMethods = msgspec.field(default_factory=lambda: GeneratorMethods())
    nicifications_path: pathlib.Path | None = None


class Config(msgspec.Struct):
    telegram_bot_api: TelegramBotAPI = msgspec.field(name="telegram-bot-api")
    generator: Generator


__all__ = (
    "Config",
    "Generator",
    "GeneratorAPI",
    "TelegramBotAPI",
    "TypedDefaultParameter",
)



import secrets
import typing

from kungfu.library.misc import from_optional
from kungfu.library.monad.option import Option

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.view.base import View

type HasherWithData[Event: BaseCute, Data] = tuple[Hasher[Event, Data], View, Data]

Event = typing.TypeVar("Event", bound=BaseCute, covariant=True)
Data = typing.TypeVar("Data", covariant=True)


def ECHO[T](__x: T) -> T:  # noqa
    return __x


class Hasher(typing.Generic[Event, Data]):
    def __init__(
        self,
        get_hash_from_data: typing.Callable[[Data], typing.Hashable | None] | None = None,
        get_data_from_event: typing.Callable[[Event], Data | None] | None = None,
    ) -> None:
        self.code = secrets.token_hex(8)
        self._get_hash_from_data = get_hash_from_data
        self._get_data_from_event = get_data_from_event

    def __call__[D](
        self: Hasher[Event, D],
        view: View,
        data: D,
        /,
    ) -> HasherWithData[Event, D]:
        return (self, view, data)

    def __hash__(self) -> int:
        return hash(self.code)

    def __repr__(self) -> str:
        return f"<Hasher {self.code}>"

    def get_hash_from_data[D](self: Hasher[Event, D], data: D) -> Option[typing.Hashable]:
        if self._get_hash_from_data is None:
            raise NotImplementedError
        return from_optional(self._get_hash_from_data(data))

    def get_data_from_event[E: BaseCute](self: Hasher[E, Data], event: E) -> Option[Data]:
        if not self._get_data_from_event:
            raise NotImplementedError
        return from_optional(self._get_data_from_event(event))

    def get_hash_from_data_from_event[E: BaseCute](
        self: Hasher[E, Data],
        event: E,
    ) -> Option[typing.Hashable]:
        return self.get_data_from_event(event).then(self.get_hash_from_data)


__all__ = ("Hasher",)
